import numpy as np
import numpy.linalg as la
from ase.units import Hartree

from gpaw.lcao.overlap import NewTwoCenterIntegrals as TwoCenterIntegrals
from gpaw.lfc import BasisFunctions
from gpaw.utilities import unpack
from gpaw.utilities.tools import dagger, lowdin, tri2full
from gpaw.lcao.tools import basis_subset2


def dots(*args):
    x = args[0]
    for M in args[1:]:
        x = np.dot(x, M)
    return x


def normalize(U, U2=None, norms=None):
    if norms is None:
        norms2 = np.ones(U.shape[1])
    else:
        norms2 = np.sqrt(norms)

    if U2 is None:
        for n, col in zip(norms2, U.T):
            col *= n / la.norm(col)
    else:
        for n, col1, col2 in zip(norms2, U.T, U2.T):
            norm = np.sqrt(np.vdot(col1, col1) + np.vdot(col2, col2))
            col1 *= n / norm
            col2 *= n / norm


def normalize2(C, S):
    C /= np.sqrt(dots(dagger(C), S, C).diagonal())


def get_rot(F_MM, V_oM, L):
    eps_M, U_MM = la.eigh(F_MM)
    indices = eps_M.real.argsort()[-L:]
    U_Ml = U_MM[:, indices]
    U_Ml /= np.sqrt(dots(U_Ml.T.conj(), F_MM, U_Ml).diagonal())

    U_ow = V_oM.copy()
    U_lw = np.dot(U_Ml.T.conj(), F_MM)
    for col1, col2 in zip(U_ow.T, U_lw.T):
        norm = np.sqrt(np.vdot(col1, col1) + np.vdot(col2, col2))
        col1 /= norm
        col2 /= norm
    return U_ow, U_lw, U_Ml


def condition_number(S):
    eps = la.eigvalsh(S).real
    return eps.max() / eps.min()


def eigvals(H, S):
    return np.sort(la.eigvals(la.solve(S, H)).real)


def get_bfs(calc):
    wfs = calc.wfs
    bfs = BasisFunctions(wfs.gd,
                         [setup.basis_functions_J for setup in wfs.setups],
                         wfs.kd, cut=True)
    bfs.set_positions(wfs.spos_ac)
    return bfs


def get_lcao_projections_HSP(calc, bfs=None, spin=0, projectionsonly=True):
    """Some title.

    if projectionsonly is True, return the projections::

      V_qnM = <psi_qn | Phi_qM>

    else, also return the Hamiltonian, overlap, and projector overlaps::

      H_qMM  = <Phi_qM| H |Phi_qM'>
      S_qMM  = <Phi_qM|Phi_qM'>
      P_aqMi = <pt^a_qi|Phi_qM>
    """
    if calc.wfs.kd.comm.size != 1:
        raise NotImplementedError('Parallelization over spin/kpt not '
                                  'implemented yet.')
    comm = calc.wfs.gd.comm
    nq = len(calc.wfs.kd.ibzk_qc)
    nao = calc.wfs.setups.nao
    dtype = calc.wfs.dtype
    if bfs is None:
        bfs = get_bfs(calc)
    tci = TwoCenterIntegrals(calc.wfs.gd.cell_cv,
                             calc.wfs.gd.pbc_c,
                             calc.wfs.setups,
                             calc.wfs.kd.ibzk_qc,
                             calc.wfs.kd.gamma)

    # Calculate projector overlaps, and (lower triangle of-) S and T matrices
    S_qMM = np.zeros((nq, nao, nao), dtype)
    T_qMM = np.zeros((nq, nao, nao), dtype)
    P_aqMi = {}

    for a in bfs.my_atom_indices:
        ni = calc.wfs.setups[a].ni
        P_aqMi[a] = np.zeros((nq, nao, ni), dtype)
    tci.calculate(calc.spos_ac, S_qMM, T_qMM, P_aqMi)

    from gpaw.lcao.atomic_correction import DenseAtomicCorrection
    dS_aii = {a: calc.wfs.setups[a].dO_ii for a in bfs.my_atom_indices}
    atomic_correction = DenseAtomicCorrection(P_aqMi, dS_aii, 0, nao)
    atomic_correction.add_overlap_correction(S_qMM)
    # calc.wfs.P_aqMi = P_aqMi
    # for q, S_MM in enumerate(S_qMM):
    #     atomic_correction.calculate_manual(wfs, q,
    # atomic_correction.add_overlap_correction(calc.wfs, S_qMM)
    calc.wfs.gd.comm.sum(S_qMM)
    calc.wfs.gd.comm.sum(T_qMM)

    # Calculate projections
    V_qnM = np.zeros((nq, calc.wfs.bd.nbands, nao), dtype)
    for kpt in calc.wfs.kpt_u:
        if kpt.s != spin:
            continue
        V_nM = V_qnM[kpt.q]
        # bfs.integrate2(kpt.psit_nG[:], V_nM, kpt.q) # all bands to save time
        for n, V_M in enumerate(V_nM):  # band-by-band to save memory
            bfs.integrate2(kpt.psit_nG[n][:], V_M, kpt.q)
        for a, P_ni in kpt.projections.items():
            dS_ii = calc.wfs.setups[a].dO_ii
            P_Mi = P_aqMi[a][kpt.q]
            V_nM += np.dot(P_ni, np.inner(dS_ii, P_Mi).conj())
    comm.sum(V_qnM)
    if projectionsonly:
        return V_qnM

    # Determine potential matrix
    vt_G = calc.hamiltonian.vt_sG[spin]
    H_qMM = np.zeros((nq, nao, nao), dtype)
    for q, H_MM in enumerate(H_qMM):
        bfs.calculate_potential_matrix(vt_G, H_MM, q)

    # Make Hamiltonian as sum of kinetic (T) and potential (V) matrices
    # and add atomic corrections
    for a, P_qMi in P_aqMi.items():
        dH_ii = unpack(calc.hamiltonian.dH_asp[a][spin])
        for P_Mi, H_MM in zip(P_qMi, H_qMM):
            H_MM += np.dot(P_Mi, np.inner(dH_ii, P_Mi).conj())
    comm.sum(H_qMM)
    H_qMM += T_qMM

    # Fill in the upper triangles of H and S
    for H_MM, S_MM in zip(H_qMM, S_qMM):
        tri2full(H_MM)
        tri2full(S_MM)
    H_qMM *= Hartree

    return V_qnM, H_qMM, S_qMM, P_aqMi


def convert_projection_data(symbols, V_qnM, H_qMM, S_qMM, P_aqMi,
                            originaltype='dzp', newtype='sz'):
    Nq = len(H_qMM)
    mask = basis_subset2(symbols, originaltype, newtype)
    mask2 = np.outer(mask, mask)
    NMnew = len(mask.nonzero()[0])
    V_qnM = V_qnM[..., mask]
    H_qMM = H_qMM[:, mask2].reshape(Nq, NMnew, NMnew)
    S_qMM = S_qMM[:, mask2].reshape(Nq, NMnew, NMnew)
    P2_aqMi = {}
    for a, P_qMi in P_aqMi.items():
        P2_aqMi[a] = P_qMi[:, mask, :]
    return V_qnM, H_qMM, S_qMM, P2_aqMi


def get_phs(calc, s=0):
    return get_lcao_projections_HSP(calc, bfs=None,
                                    spin=s, projectionsonly=False)


class ProjectedWannierFunctions:
    def __init__(self, projections, h_lcao, s_lcao, eigenvalues, kpoints,
                 L_k=None, M_k=None, N=None, fixedenergy=None):
        """projections[n,i] = <psi_n|f_i>
           h_lcao[i1, i2] = <f_i1|h|f_i2>
           s_lcao[[i1, i2] = <f_i1|f_i2>
           eps_n: Exact eigenvalues
           L: Number of extra degrees of freedom
           M: Number of states to exactly span
           N: Total number of bands in the calculation

           Methods:
           -- get_hamiltonian_and_overlap_matrix --
           will return the hamiltonian and identity operator
           in the projected wannier function basis.
           The following steps are performed:

           1) calculate_edf       -> self.b_il
           2) calculate_rotations -> self.Uo_mi and self.Uu_li
           3) calculate_overlaps  -> self.S_ii
           4) calculate_hamiltonian_matrix -> self.H_ii

           -- get_eigenvalues --
           gives the eigenvalues of of the hamiltonian in the
           projected wannier function basis.

           -- indices --
           i localized function index
           n eigenstate index
           l edf index
           m fixed eigenstate index
           k k-point index
           """

        self.eps_kn = eigenvalues
        self.ibzk_kc = kpoints
        self.nk = len(self.ibzk_kc)
        self.V_kni = projections  # <psi_n1|f_i1>
        self.dtype = self.V_kni.dtype
        self.Nw = self.V_kni.shape[2]
        self.s_lcao_kii = s_lcao  # F_ii[i1,i2] = <f_i1|f_i2>
        self.h_lcao_kii = h_lcao

        if N is None:
            N = self.V_kni.shape[1]
        self.N = N

        if fixedenergy is None:
            raise NotImplementedError(
                'Only fixedenergy is implemented for now')
        else:
            self.fixedenergy = fixedenergy
            self.M_k = [sum(eps_n <= fixedenergy) for eps_n in self.eps_kn]
            self.L_k = [self.Nw - M for M in self.M_k]
            print("fixedenergy =", self.fixedenergy)

        print('N =', self.N)
        print('skpt_kc = ')
        print(self.ibzk_kc)
        print('M_k =', self.M_k)
        print('L_k =', self.L_k)
        print('Nw =', self.Nw)

    def get_hamiltonian_and_overlap_matrix(self, useibl=True):
        self.calculate_edf(useibl=useibl)
        self.calculate_rotations()
        self.calculate_overlaps()
        self.calculate_hamiltonian_matrix(useibl=useibl)
        return self.H_kii, self.S_kii

    def calculate_edf(self, useibl=True):
        """Calculate the coefficients b_il in the expansion of the EDF.

        ``|phi_l> = sum_i b_il |f^u_i>``, in terms of ``|f^u_i> = P^u|f_i>``.

        To use the infinite band limit set useibl=True.
        N is the total number of bands to use.
        """

        for k, L in enumerate(self.L_k):
            if L == 0:
                assert L != 0, 'L_k=0 for k=%i. Not implemented' % k

        self.Vo_kni = [V_ni[:M] for V_ni, M in zip(self.V_kni, self.M_k)]

        self.Fo_kii = np.asarray([np.dot(dagger(Vo_ni), Vo_ni)
                                  for Vo_ni in self.Vo_kni])

        if useibl:
            self.Fu_kii = self.s_lcao_kii - self.Fo_kii
        else:
            self.Vu_kni = [V_ni[M:self.N]
                           for V_ni, M in zip(self.V_kni, self.M_k)]
            self.Fu_kii = np.asarray([np.dot(dagger(Vu_ni), Vu_ni)
                                     for Vu_ni in self.Vu_kni])
        self.b_kil = []
        for Fu_ii, L in zip(self.Fu_kii, self.L_k):
            b_i, b_ii = la.eigh(Fu_ii)
            ls = b_i.real.argsort()[-L:]
            b_il = b_ii[:, ls]  # pick out the eigenvec with largest eigenvals.
            normalize2(b_il, Fu_ii)  # normalize the EDF: <phi_l|phi_l> = 1
            self.b_kil.append(b_il)

    def calculate_rotations(self):
        Uo_kni = [Vo_ni.copy() for Vo_ni in self.Vo_kni]
        Uu_kli = [np.dot(dagger(b_il), Fu_ii)
                  for b_il, Fu_ii in zip(self.b_kil, self.Fu_kii)]
        # Normalize such that <omega_i|omega_i> = <f_i|f_i>
        for Uo_ni, Uu_li, s_ii in zip(Uo_kni, Uu_kli, self.s_lcao_kii):
            normalize(Uo_ni, Uu_li, s_ii.diagonal())
        self.Uo_kni = Uo_kni
        self.Uu_kli = Uu_kli

    def calculate_overlaps(self):
        Wo_kii = [np.dot(dagger(Uo_ni), Uo_ni) for Uo_ni in self.Uo_kni]
        Wu_kii = [np.dot(dagger(Uu_li), Uu_li) for Uu_li in self.Uu_kli]
        # Wu_kii = [dots(dagger(Uu_li), dagger(b_il), Fu_ii, b_il, Uu_li)
        # for Uu_li, b_il, Fu_ii in zip(self.Uu_kli, self.b_kil, self.Fu_kii)]
        self.S_kii = np.asarray(Wo_kii) + np.asarray(Wu_kii)

    def get_condition_number(self):
        return np.asarray([condition_number(S) for S in self.S_kii])

    def calculate_hamiltonian_matrix(self, useibl=True):
        """Calculate H_kij = H^o_i(k)j(k) + H^u_i(k)j(k)
           i(k): Bloch sum of omega_i
        """

        epso_kn = [eps_n[:M] for eps_n, M in zip(self.eps_kn, self.M_k)]
        self.Ho_kii = np.asarray([np.dot(dagger(Uo_ni) * epso_n, Uo_ni)
                                  for Uo_ni, epso_n in zip(self.Uo_kni,
                                                           epso_kn)])

        if self.h_lcao_kii is not None and useibl:
            print("Using h_lcao and infinite band limit")
            Huf_kii = [h_lcao_ii - np.dot(dagger(Vo_ni) * epso_n, Vo_ni)
                       for h_lcao_ii, Vo_ni, epso_n in zip(self.h_lcao_kii,
                                                           self.Vo_kni,
                                                           epso_kn)]
            self.Huf_kii = np.asarray(Huf_kii)
        else:
            print("Using finite band limit (not using h_lcao)")
            epsu_kn = [eps_n[M:self.N]
                       for eps_n, M in zip(self.eps_kn, self.M_k)]
            Huf_kii = [np.dot(dagger(Vu_ni) * epsu_n, Vu_ni)
                       for Vu_ni, epsu_n in zip(self.Vu_kni, epsu_kn)]
            self.Huf_kii = np.asarray(Huf_kii)

        Hu_kii = [dots(dagger(Uu_li), dagger(b_il), Huf_ii, b_il, Uu_li)
                  for Uu_li, b_il, Huf_ii in zip(self.Uu_kli, self.b_kil,
                                                 self.Huf_kii)]
        self.Hu_kii = np.asarray(Hu_kii)
        self.H_kii = self.Ho_kii + self.Hu_kii

    def get_eigenvalues(self):
        return np.asarray([eigvals(H, S)
                           for H, S in zip(self.H_kii, self.S_kii)])

    def get_lcao_eigenvalues(self):
        return np.asarray([eigvals(H, S)
                           for H, S in zip(self.h_lcao_kii, self.s_lcao_kii)])

    def get_norm_of_projection(self):
        Sinv_kii = np.asarray([la.inv(S_ii) for S_ii in self.S_kii])

        normo_kn = np.asarray([dots(Uo_ni, Sinv_ii, dagger(Uo_ni)).diagonal()
                               for Uo_ni, Sinv_ii in
                               zip(self.Uo_kni, Sinv_kii)])

        Vu_kni = np.asarray([V_ni[M:self.N]
                             for V_ni, M in zip(self.V_kni, self.M_k)])

        Pu_kni = [dots(Vu_ni, b_il, Uu_li)
                  for Vu_ni, b_il, Uu_li in
                  zip(Vu_kni, self.b_kil, self.Uu_kli)]

        normu_kn = np.asarray([dots(Pu_ni, Sinv_ii, dagger(Pu_ni)).diagonal()
                               for Pu_ni, Sinv_ii in zip(Pu_kni, Sinv_kii)])

        return np.hstack((normo_kn, normu_kn))

    def calculate_functions(self, calc, basis, k=0):
        from gpaw.io import FileReference
        psit_nG = calc.wfs.kpt_u[k].psit_nG
        Uo_ni = self.Uo_kni[k]
        tarinstance = isinstance(psit_nG, FileReference)
        if tarinstance:
            psit_nG = np.asarray([psit_nG[i] for i in range(self.M_k[k])])

        basis_functions = get_bfs(calc)
        b_il, Uu_li, Vo_ni = self.b_kil[k], self.Uu_kli[k], self.Vo_kni[k]
        a_iG = -np.tensordot(Vo_ni, psit_nG, axes=[0, 0])  # a_iG = -fo_iG
        # self.fo_iG = -a_iG.copy()#f projected onto the occupied subspace
        C_iM = np.identity(self.Nw, dtype=self.dtype)
        basis_functions.lcao_to_grid(C_iM, a_iG, q=-1)  # a_iG=fu_iG=f_iG-fo_iG
        # self.f_iG = self.fo_iG + a_iG # check
        a_iG = np.tensordot(b_il, a_iG, axes=[0, 0])  # a_iG = EDF
        a_iG = np.tensordot(Uu_li, a_iG, axes=[0, 0])  # a_iG = wu_iG
        a_iG += np.tensordot(Uo_ni, psit_nG, axes=[0, 0])  # ai_G = wu_iG+wo_iG
        self.w_iG = a_iG

    def get_mlwf_initial_guess(self):
        """calculate initial guess for maximally localized
        wannier functions. Does not work for the infinite band limit.
        cu_nl: rotation coefficents of unoccupied states
        U_ii: rotation matrix of eigenstates and edf.
        """
        Vu_ni = self.Vu_ni[self.M: self.N]
        cu_nl = np.dot(Vu_ni, self.b_il)
        U_ii = np.vstack((self.Uo_ni, self.Uu_li))
        lowdin(U_ii)
        return U_ii, cu_nl



"""Excitation lists base classes

"""
from math import sqrt
import numpy as np

from ase.units import Ha

import gpaw.mpi as mpi
from gpaw.io.logger import GPAWLogger


def get_filehandle(cls, filename, mode='r'):
    cls.filename = filename
    if filename.endswith('.gz'):
        try:
            import gzip
            return gzip.open(filename, mode + 't')
        except ModuleNotFoundError:
            pass
    return open(filename, mode)


class ExcitationLogger(GPAWLogger):
    def header(self):
        pass


class ExcitationList(list):
    """General Excitation List class.
    """
    def __init__(self, log=None, txt='-'):
        # initialise empty list
        list.__init__(self)
        self.energy_to_eV_scale = Ha

        # set output
        if log is not None:
            self.log = log
        else:
            self.log = ExcitationLogger(world=mpi.world)
            self.log.fd = txt

    @property
    def calc(self):
        1 / 0

    def get_energies(self):
        """Get excitation energies in Hartrees"""
        el = []
        for ex in self:
            el.append(ex.get_energy())
        return np.array(el)

    def get_trk(self):
        """Evaluate the Thomas Reiche Kuhn sum rule"""
        trkm = np.zeros((3))
        for ex in self:
            me = ex.get_dipole_me()
            trkm += ex.get_energy() * (me.real ** 2 + me.imag ** 2)
        return 2. * trkm  # scale to get the number of electrons XXX spinpol ?

    def get_polarizabilities(self, lmax=7):
        """Calculate the Polarisabilities
        see Jamorski et al. J. Chem. Phys. 104 (1996) 5134"""
        S = np.zeros((lmax + 1))
        for ex in self:
            e = ex.get_energy()
            f = ex.get_oscillator_strength()[0]
            for l in range(lmax + 1):
                S[l] += e ** (-2 * l) * f
        return S

    def __truediv__(self, x):
        return self.__mul__(1. / x)

    __div__ = __truediv__

    def __rmul__(self, x):
        return self.__mul__(x)

    def __mul__(self, x):
        """Multiply with a number"""
        if isinstance(x, (float, int)):
            result = self.__class__()
            result.dtype = self.dtype
            for kss in self:
                result.append(x * kss)
            return result
        else:
            return RuntimeError('not a number')

    def __sub__(self, other):
        result = self.__class__()
        result.dtype = self.dtype
        assert len(self) == len(other)
        for kss, ksso in zip(self, other):
            result.append(kss - ksso)
        return result

    def __str__(self):
        string = '# ' + str(type(self))
        if len(self) != 0:
            string += ', %d excitations:' % len(self)
        string += '\n'
        for ex in self:
            string += '#  ' + ex.__str__() + '\n'
        return string


class Excitation:

    def get_energy(self):
        """Get the excitations energy relative to the ground state energy
        in Hartrees.
        """
        return self.energy

    def get_dipole_me(self, form='r'):
        """return the excitations dipole matrix element
        including the occupation factor sqrt(fij)"""
        if form == 'r':
            # length form
            return self.me / sqrt(self.energy)
        elif form == 'v':
            # velocity form
            return - np.sqrt(self.fij) * self.muv
        else:
            raise RuntimeError('Unknown form >' + form + '<')

    def get_dipole_tensor(self, form='r'):
        """Return the "oscillator strength tensor"

        self.me is assumed to be::

          form='r': sqrt(f * E) * <I|r|J>,
          form='v': sqrt(f / E) * <I|d/(dr)|J>

        for f = multiplicity, E = transition energy and initial and
        final states::

          |I>, |J>
        """

        if form == 'r':
            # length form
            me = self.me
        elif form == 'v':
            # velocity form
            me = self.muv * np.sqrt(self.fij * self.energy)
        else:
            raise RuntimeError('Unknown form >' + form + '<')

        return 2 * np.outer(me, me.conj())

    def get_oscillator_strength(self, form='r'):
        """Return the excitations dipole oscillator strength."""
        me2_c = self.get_dipole_tensor(form).diagonal().real
        return np.array([np.sum(me2_c) / 3.] + me2_c.tolist())

    def get_rotatory_strength(self, form='r', units='cgs'):
        """Return rotatory strength"""
        if self.magn is None:
            raise RuntimeError('Magnetic moment not available.')

        if units == 'cgs':
            # 10^-40 esu cm erg / G
            # = 3.33564095 * 10^-15 A^2 m^3 s
            # conversion factor after
            # T. B. Pedersen and A. E. Hansen,
            # Chem. Phys. Lett. 246 (1995) 1
            # pre = 471.43
            # From TurboMole
            pre = 64604.8164
        elif units == 'a.u.':
            pre = 1.
        else:
            raise RuntimeError('Unknown units >' + units + '<')

        if form == 'r':
            # length form
            mu = self.mur
        elif form == 'v':
            # velocity form
            mu = self.muv
        else:
            raise RuntimeError('Unknown form >' + form + '<')

        return -pre * np.dot(mu, self.magn)

    def set_energy(self, E):
        """Set the excitations energy relative to the ground state energy"""
        self.energy = E



import numpy as np

import pytest

from gpaw.mpi import world
from gpaw.response.pw_parallelization import Blocks1D


@pytest.mark.response
def test_blocks1d_collect():
    """Test the ability to collect an array distributed over the first
    dimension."""
    dat_i = np.arange(150)
    dat_ij = dat_i.reshape((10, 15))
    dat_ijk = dat_i.reshape((5, 3, 10))

    for array in [dat_i, dat_ij, dat_ijk]:
        blocks = Blocks1D(world, array.shape[0])
        local_array = array[blocks.myslice]

        # Test all-gather
        collected_array = blocks.all_gather(local_array)
        assert np.all(array == collected_array)

        # Test gather
        collected_array = blocks.gather(local_array)
        if world.rank == 0:
            assert np.all(array == collected_array)
        else:
            assert collected_array is None



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.
"""Main gpaw module."""
import os
import sys
import contextlib
from pathlib import Path
from typing import List, Dict, Union, Any, TYPE_CHECKING

__version__ = '23.6.0'
__ase_version_required__ = '3.22.1'
__all__ = ['GPAW',
           'Mixer', 'MixerSum', 'MixerDif', 'MixerSum2',
           'MixerFull',
           'CG', 'Davidson', 'RMMDIIS', 'DirectLCAO',
           'PoissonSolver',
           'FermiDirac', 'MethfesselPaxton', 'MarzariVanderbilt',
           'PW', 'LCAO', 'FD',
           'restart']

setup_paths: List[Union[str, Path]] = []
is_gpaw_python = '_gpaw' in sys.builtin_module_names
dry_run = 0

# When type-checking or running pytest, we want the debug-wrappers enabled:
debug: bool = (TYPE_CHECKING or
               'pytest' in sys.modules or
               bool(sys.flags.debug))


@contextlib.contextmanager
def disable_dry_run():
    """Context manager for temporarily disabling dry-run mode.

    Useful for skipping exit in the GPAW constructor.
    """
    global dry_run
    size = dry_run
    dry_run = 0
    yield
    dry_run = size


if 'OMP_NUM_THREADS' not in os.environ:
    os.environ['OMP_NUM_THREADS'] = '1'


from gpaw.broadcast_imports import broadcast_imports  # noqa

with broadcast_imports:
    import os
    import runpy
    import warnings
    from argparse import ArgumentParser, REMAINDER, RawDescriptionHelpFormatter

    # With gpaw-python BLAS symbols are in global scope and we need to
    # ensure that NumPy and SciPy use symbols from their own dependencies
    if is_gpaw_python:
        old_dlopen_flags = sys.getdlopenflags()
        sys.setdlopenflags(old_dlopen_flags | os.RTLD_DEEPBIND)
    import numpy as np
    import scipy.linalg  # noqa: F401
    if is_gpaw_python:
        sys.setdlopenflags(old_dlopen_flags)
    import _gpaw

SCIPY_VERSION = [int(x) for x in scipy.__version__.split('.')[:2]]

if getattr(_gpaw, 'version', 0) != 3:
    raise ImportError('Please recompile GPAW''s C-extensions!')


class ConvergenceError(Exception):
    pass


class KohnShamConvergenceError(ConvergenceError):
    pass


class PoissonConvergenceError(ConvergenceError):
    pass


class KPointError(Exception):
    pass


class BadParallelization(Exception):
    """Error indicating missing parallelization support."""
    pass


libraries: Dict[str, str] = {}
if hasattr(_gpaw, 'lxcXCFunctional'):
    libraries['libxc'] = getattr(_gpaw, 'libxc_version', '2.x.y')
else:
    libraries['libxc'] = ''


def parse_arguments(argv):
    version = sys.version.replace('\n', '')
    p = ArgumentParser(usage='%(prog)s [OPTION ...] [-c | -m] SCRIPT'
                       ' [ARG ...]',
                       description='Run a parallel GPAW calculation.\n\n'
                       f'Compiled with:\n  Python {version}',
                       formatter_class=RawDescriptionHelpFormatter)

    p.add_argument('--command', '-c', action='store_true',
                   help='execute Python string given as SCRIPT')
    p.add_argument('--module', '-m', action='store_true',
                   help='run library module given as SCRIPT')
    p.add_argument('-W', metavar='argument',
                   action='append', default=[], dest='warnings',
                   help='warning control.  See the documentation of -W for '
                   'the Python interpreter')
    p.add_argument('script', metavar='SCRIPT',
                   help='calculation script')
    p.add_argument('options', metavar='ARG',
                   help='arguments forwarded to SCRIPT', nargs=REMAINDER)

    args = p.parse_args(argv[1:])

    if args.command and args.module:
        p.error('-c and -m are mutually exclusive')

    sys.argv = [args.script] + args.options

    for w in args.warnings:
        # Need to convert between python -W syntax to call
        # warnings.filterwarnings():
        warn_args = w.split(':')
        assert len(warn_args) <= 5

        if warn_args[0] == 'all':
            warn_args[0] = 'always'
        if len(warn_args) >= 3:
            # e.g. 'UserWarning' (string) -> UserWarning (class)
            warn_args[2] = globals().get(warn_args[2])
        if len(warn_args) == 5:
            warn_args[4] = int(warn_args[4])

        warnings.filterwarnings(*warn_args, append=True)

    return args


def main():
    gpaw_args = parse_arguments(sys.argv)
    # The normal Python interpreter puts . in sys.path, so we also do that:
    sys.path.insert(0, '.')
    # Stacktraces can be shortened by running script with
    # PyExec_AnyFile and friends.  Might be nicer
    if gpaw_args.command:
        d = {'__name__': '__main__'}
        exec(gpaw_args.script, d, d)
    elif gpaw_args.module:
        # Python has: python [-m MOD] [-c CMD] [SCRIPT]
        # We use a much better way: gpaw-python [-m | -c] SCRIPT
        runpy.run_module(gpaw_args.script, run_name='__main__')
    else:
        runpy.run_path(gpaw_args.script, run_name='__main__')


if debug:
    np.seterr(over='raise', divide='raise', invalid='raise', under='ignore')
    oldempty = np.empty
    oldempty_like = np.empty_like

    def empty(*args, **kwargs):
        a = oldempty(*args, **kwargs)
        try:
            a.fill(np.nan)
        except ValueError:
            a.fill(42)
        return a

    def empty_like(*args, **kwargs):
        a = oldempty_like(*args, **kwargs)
        try:
            a.fill(np.nan)
        except ValueError:
            a.fill(-42)
        return a

    np.empty = empty
    np.empty_like = empty_like


with broadcast_imports:
    from gpaw.calculator import GPAW as OldGPAW
    from gpaw.mixer import Mixer, MixerSum, MixerDif, MixerSum2, \
        MixerFull
    from gpaw.eigensolvers import Davidson, RMMDIIS, CG, DirectLCAO
    from gpaw.poisson import PoissonSolver
    from gpaw.occupations import (FermiDirac, MethfesselPaxton,
                                  MarzariVanderbilt)
    from gpaw.wavefunctions.lcao import LCAO
    from gpaw.wavefunctions.pw import PW
    from gpaw.wavefunctions.fd import FD


def GPAW(*args, **kwargs) -> Any:
    if os.environ.get('GPAW_NEW'):
        from gpaw.new.ase_interface import GPAW as NewGPAW
        return NewGPAW(*args, **kwargs)
    return OldGPAW(*args, **kwargs)


def restart(filename, Class=GPAW, **kwargs):
    calc = Class(filename, **kwargs)
    atoms = calc.get_atoms()
    return atoms, calc


def read_rc_file():
    home = os.environ.get('HOME')
    if home is not None:
        rc = os.path.join(home, '.gpaw', 'rc.py')
        if os.path.isfile(rc):
            # Read file in ~/.gpaw/rc.py
            with open(rc) as fd:
                exec(fd.read())


def initialize_data_paths():
    try:
        setup_paths[:0] = os.environ['GPAW_SETUP_PATH'].split(os.pathsep)
    except KeyError:
        if len(setup_paths) == 0:
            if os.pathsep == ';':
                setup_paths[:] = [r'C:\gpaw-setups']
            else:
                setup_paths[:] = ['/usr/local/share/gpaw-setups',
                                  '/usr/share/gpaw-setups']


read_rc_file()
initialize_data_paths()


def RMM_DIIS(*args, **kwargs):
    warnings.warn('Please use RMMDIIS instead of RMM_DIIS')
    return RMMDIIS(*args, **kwargs)



import numpy as np
import pytest

from gpaw.elph import ResonantRamanCalculator


def make_hermitian(mat):
    n = mat.shape[0]
    for i in range(n):
        for j in range(i + 1, n):
            mat[j, i] = mat[i, j].conj()


@pytest.mark.serial
def test_resonant_term():
    rt = ResonantRamanCalculator.resonant_term

    f_n = np.array([1.0, 0.0, 0.0])
    f_vc = np.outer(f_n, 1.0 - f_n)
    assert f_vc[0] == pytest.approx([0.0, 1.0, 1.0])
    print(f_vc)

    E_n = np.array([-1.0, 0.0, 1.0])
    E_vc = np.zeros((3, 3), dtype=complex) + 1j * 0.1
    for n in range(3):
        E_vc[n] += (E_n - E_n[n])
    assert E_vc[0] == pytest.approx([0.0 + 1j * 0.1,
                                     1.0 + 1j * 0.1,
                                     2.0 + 1j * 0.1])

    nn = np.zeros((3, 3), dtype=complex)
    nn[0, 0] = 1.0
    nn[1, 1] = 2.0
    nn[2, 2] = 3.0
    nn[0, 1] = 0.5 + 0.1j
    nn[0, 2] = 0.25 + 0.1j
    nn[1, 2] = 0.125 + 0.1j
    make_hermitian(nn)
    # print(nn)
    mom_dnn = np.zeros((2, 3, 3), dtype=complex)
    mom_dnn[0] = nn
    mom_dnn[1] = 2. * nn

    elph_lnn = np.zeros((2, 3, 3), dtype=complex)
    elph_lnn[1] = 3. * nn

    wph_w = np.array([0.0, 0.2])

    term = rt(f_vc, E_vc, mom_dnn, elph_lnn, 0, 3, 1.0, wph_w)
    # we'll want to check whether this is the expect number
    assert term[1] == pytest.approx(-29.768613861386154 - 64.0461386138614j)

    # complement polarisations need to yield the same result
    term2 = rt(f_vc, E_vc, mom_dnn[[1, 0], :, :], elph_lnn, 0, 3, 1.0, wph_w)
    assert term[1] == pytest.approx(term2[1])



import pytest
from ase.build import molecule

from gpaw import GPAW, KohnShamConvergenceError
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.tddft import TDDFT
from gpaw.utilities.timelimit import TimeLimiter


@pytest.mark.later
def test_timelimit(in_tmp_dir):
    # Atoms
    atoms = molecule('Na2')
    atoms.center(vacuum=4.0)

    # Ground-state calculation that will never converge
    maxiter = 10
    calc = GPAW(mode='lcao', basis='sz(dzp)', setups='1', nbands=1,
                convergence={'density': 1e-100},
                maxiter=maxiter)
    atoms.calc = calc

    tl = TimeLimiter(calc, timelimit=0, output='scf.txt')
    tl.reset('scf', min_updates=3)
    try:
        atoms.get_potential_energy()
    except KohnShamConvergenceError:
        assert calc.scf.maxiter < maxiter, 'TimeLimiter did not break SCF loop'
    else:
        raise AssertionError('SCF loop ended too early')
    calc.write('gs.gpw', mode='all')

    # LCAOTDDFT calculation that will never finish
    td_calc = LCAOTDDFT('gs.gpw')
    tl = TimeLimiter(td_calc, timelimit=0, output='lcaotddft.txt')
    tl.reset('tddft', min_updates=3)
    td_calc.propagate(10, maxiter - td_calc.niter)
    assert td_calc.maxiter < maxiter, 'TimeLimiter did not break TDDFT loop'

    # Test mode='fd'

    # Prepare ground state
    calc = GPAW(mode='fd', setups='1', maxiter=1, nbands=1)
    atoms.calc = calc
    try:
        atoms.get_potential_energy()
    except KohnShamConvergenceError:
        pass
    calc.write('gs.gpw', mode='all')

    # TDDFT calculation that will never finish
    td_calc = TDDFT('gs.gpw')
    tl = TimeLimiter(td_calc, timelimit=0, output='tddft.txt')
    tl.reset('tddft', min_updates=3)
    td_calc.propagate(10, maxiter - td_calc.niter)
    assert td_calc.maxiter < maxiter, 'TimeLimiter did not break TDDFT loop'



import os
import warnings
from contextlib import contextmanager
from pathlib import Path

import numpy as np
import pytest
from _pytest.tmpdir import _mk_tmp
from ase import Atoms
from ase.build import bulk
from ase.lattice.hexagonal import Graphene
from ase.io import read
from gpaw import GPAW, PW, Davidson, FermiDirac, setup_paths
from gpaw.cli.info import info
from gpaw.mpi import broadcast, world
from gpaw.utilities import devnull
from ase.lattice.compounds import L1_2
from gpaw import Mixer


@contextmanager
def execute_in_tmp_path(request, tmp_path_factory):
    if world.rank == 0:
        # Obtain basename as
        # * request.function.__name__  for function fixture
        # * request.module.__name__    for module fixture
        basename = getattr(request, request.scope).__name__
        path = tmp_path_factory.mktemp(basename)
    else:
        path = None
    path = broadcast(path)
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(cwd)


@pytest.fixture(scope='function')
def in_tmp_dir(request, tmp_path_factory):
    """Run test function in a temporary directory."""
    with execute_in_tmp_path(request, tmp_path_factory) as path:
        yield path


@pytest.fixture(scope='module')
def module_tmp_path(request, tmp_path_factory):
    """Run test module in a temporary directory."""
    with execute_in_tmp_path(request, tmp_path_factory) as path:
        yield path


@pytest.fixture
def add_cwd_to_setup_paths():
    """Temporarily add current working directory to setup_paths."""
    try:
        setup_paths[:0] = ['.']
        yield
    finally:
        del setup_paths[:1]


response_band_cutoff = dict(
)


def with_band_cutoff(*, gpw, band_cutoff):
    # Store the band cutoffs in a dictionary to aid response tests
    response_band_cutoff[gpw] = band_cutoff

    def decorator(func):
        def wrapper(*args, **kwargs):
            return func(*args, band_cutoff=band_cutoff, **kwargs)
        return wrapper

    return decorator


@pytest.fixture(scope='session')
def gpw_files(request, tmp_path_factory):
    """Reuse gpw-files.

    Returns a dict mapping names to paths to gpw-files.  If you
    want to reuse gpw-files from an earlier pytest session then set the
    ``$GPW_TEST_FILES`` environment variable and the files will be written
    to that folder.

    Example::

        def test_something(gpw_files):
            calc = GPAW(gpw_files['h2_lcao_wfs'])
            ...

    Possible systems are:

    * Bulk BCC-Li with 3x3x3 k-points: ``bcc_li_pw``, ``bcc_li_fd``,
      ``bcc_li_lcao``.

    * O2 molecule: ``o2_pw``.

    * H2 molecule: ``h2_pw``, ``h2_fd``, ``h2_lcao``.

    * H2 molecule (not centered): ``h2_pw_0``.

    * Spin-polarized H atom: ``h_pw``.

    * Polyethylene chain.  One unit, 3 k-points, no symmetry:
      ``c2h4_pw_nosym``.  Three units: ``c6h12_pw``.

    * Bulk TiO2 with 4x4x4 k-points: ``ti2o4_pw`` and ``ti2o4_pw_nosym``.

    * Bulk BN (zinkblende) with 2x2x2 k-points and 9 converged bands:
      ``bn_pw``.

    * h-BN layer with 3x3x1 (gamma center) k-points and 26 converged bands:
      ``hbn_pw``.

    * Graphene with 6x6x1 k-points: ``graphene_pw``

    * MoS2 with 6x6x1 k-points: ``mos2_pw``

    * NiCl2 with 6x6x1 k-points: ``nicl2_pw``

    * V2Br4 (AFM monolayer), LDA, 4x2x1 k-points, 28(+1) converged bands:
      ``v2br4_pw`` and ``v2br4_pw_nosym``

    * Bulk Si, LDA, 2x2x2 k-points (gamma centered): ``si_pw``

    * Bulk Si, LDA, 4x4x4 k-points, 8(+1) converged bands: ``fancy_si_pw``
      and ``fancy_si_pw_nosym``

    * Bulk Fe, LDA, 4x4x4 k-points, 9(+1) converged bands: ``fe_pw``
      and ``fe_pw_nosym``

    * Bulk C, LDA, 2x2x2 k-points (gamma centered), ``c_pw``

    * Bulk Co (HCP), 4x4x4 k-points, 12(+1) converged bands: ``co_pw``
      and ``co_pw_nosym``

    * Bulk SrVO3 (SC), 3x3x3 k-points, 20(+1) converged bands: ``srvo3_pw``
      and ``srvo3_pw_nosym``

    * Bulk Al, LDA, 4x4x4 k-points, 10(+1) converged bands: ``al_pw``
      and ``al_pw_nosym``

    * Bulk Ag, LDA, 2x2x2 k-points, 6 converged bands,
      2eV U on d-band: ``ag_pw``

    * Bulk GaAs, LDA, 4x4x4 k-points, 8(+1) bands converged: ``gaas_pw``
      and ``gaas_pw_nosym``


    Files with wave functions are also available (add ``_wfs`` to the names).
    """
    path = os.environ.get('GPW_TEST_FILES')
    if not path:
        warnings.warn(
            'Note that you can speed up the tests by reusing gpw-files '
            'from an earlier pytest session: '
            'set the $GPW_TEST_FILES environment variable and the '
            'files will be written to/read from that folder. '
            'See: https://wiki.fysik.dtu.dk/gpaw/devel/testing.html'
            '#gpaw.test.conftest.gpw_files')
        if world.rank == 0:
            path = _mk_tmp(request, tmp_path_factory)
        else:
            path = None
        path = broadcast(path)
    return GPWFiles(Path(path))


class GPWFiles:
    """Create gpw-files."""
    def __init__(self, path: Path):
        self.path = path
        path.mkdir(exist_ok=True)
        self.gpw_files = {}
        for file in path.glob('*.gpw'):
            self.gpw_files[file.name[:-4]] = file

    def __getitem__(self, name: str) -> Path:
        if name not in self.gpw_files:
            rawname, _, _ = name.partition('_wfs')
            calc = getattr(self, rawname)()
            path = self.path / (rawname + '.gpw')
            calc.write(path)
            self.gpw_files[rawname] = path
            path = self.path / (rawname + '_wfs.gpw')
            calc.write(path, mode='all')
            self.gpw_files[rawname + '_wfs'] = path
        return self.gpw_files[name]

    def bcc_li_pw(self):
        return self.bcc_li({'name': 'pw', 'ecut': 200})

    def bcc_li_fd(self):
        return self.bcc_li({'name': 'fd'})

    def bcc_li_lcao(self):
        return self.bcc_li({'name': 'lcao'})

    def bcc_li(self, mode):
        li = bulk('Li', 'bcc', 3.49)
        li.calc = GPAW(mode=mode,
                       kpts=(3, 3, 3),
                       txt=self.path / f'bcc_li_{mode["name"]}.txt')
        li.get_potential_energy()
        return li.calc

    def h2_pw(self):
        return self.h2({'name': 'pw', 'ecut': 200})

    def h2_fd(self):
        return self.h2({'name': 'fd'})

    def h2_lcao(self):
        return self.h2({'name': 'lcao'})

    def h2(self, mode):
        h2 = Atoms('H2', positions=[[0, 0, 0], [0.74, 0, 0]])
        h2.center(vacuum=2.5)
        h2.calc = GPAW(mode=mode,
                       txt=self.path / f'h2_{mode["name"]}.txt')
        h2.get_potential_energy()
        return h2.calc

    def h2_pw_0(self):
        h2 = Atoms('H2',
                   positions=[[-0.37, 0, 0], [0.37, 0, 0]],
                   cell=[5.74, 5, 5],
                   pbc=True)
        h2.calc = GPAW(mode={'name': 'pw', 'ecut': 200},
                       txt=self.path / 'h2_pw_0.txt')
        h2.get_potential_energy()
        return h2.calc

    def h2_bcc_afm(self):
        a = 2.75
        atoms = bulk(name='H', crystalstructure='bcc', a=a, cubic=True)
        atoms.set_initial_magnetic_moments([1., -1.])

        atoms.calc = GPAW(xc='LDA',
                          txt=self.path / 'h2_bcc_afm.txt',
                          mode=PW(250),
                          nbands=4,
                          convergence={'bands': 4},
                          kpts={'density': 2.0, 'gamma': True})
        atoms.get_potential_energy()
        return atoms.calc

    def h_pw(self):
        h = Atoms('H', magmoms=[1])
        h.center(vacuum=4.0)
        h.calc = GPAW(mode={'name': 'pw', 'ecut': 500},
                      txt=self.path / 'h_pw.txt')
        h.get_potential_energy()
        return h.calc

    def o2_pw(self):
        d = 1.1
        a = Atoms('O2', positions=[[0, 0, 0], [d, 0, 0]], magmoms=[1, 1])
        a.center(vacuum=4.0)
        a.calc = GPAW(mode={'name': 'pw', 'ecut': 800},
                      txt=self.path / 'o2_pw.txt')
        a.get_potential_energy()
        return a.calc

    def Cu3Au_qna(self):
        ecut = 300
        kpts = (1, 1, 1)

        QNA = {'alpha': 2.0,
               'name': 'QNA',
               'stencil': 1,
               'orbital_dependent': False,
               'parameters': {'Au': (0.125, 0.1), 'Cu': (0.0795, 0.005)},
               'setup_name': 'PBE',
               'type': 'qna-gga'}

        atoms = L1_2(['Au', 'Cu'], latticeconstant=3.7)
        atoms[0].position[0] += 0.01  # Break symmetry already here
        calc = GPAW(mode=PW(ecut),
                    eigensolver=Davidson(2),
                    nbands='120%',
                    mixer=Mixer(0.4, 7, 50.0),
                    parallel=dict(domain=1),
                    convergence={'density': 1e-4},
                    xc=QNA,
                    kpts=kpts,
                    txt=self.path / 'Cu3Au.txt')
        atoms.calc = calc
        atoms.get_potential_energy()
        return atoms.calc

    def co_lcao(self):
        d = 1.1
        co = Atoms('CO', positions=[[0, 0, 0], [d, 0, 0]])
        co.center(vacuum=4.0)
        co.calc = GPAW(mode='lcao',
                       txt=self.path / 'co_lcao.txt')
        co.get_potential_energy()
        return co.calc

    def c2h4_pw_nosym(self):
        d = 1.54
        h = 1.1
        x = d * (2 / 3)**0.5
        z = d / 3**0.5
        pe = Atoms('C2H4',
                   positions=[[0, 0, 0],
                              [x, 0, z],
                              [0, -h * (2 / 3)**0.5, -h / 3**0.5],
                              [0, h * (2 / 3)**0.5, -h / 3**0.5],
                              [x, -h * (2 / 3)**0.5, z + h / 3**0.5],
                              [x, h * (2 / 3)**0.5, z + h / 3**0.5]],
                   cell=[2 * x, 0, 0],
                   pbc=(1, 0, 0))
        pe.center(vacuum=2.0, axis=(1, 2))
        pe.calc = GPAW(mode='pw',
                       kpts=(3, 1, 1),
                       symmetry='off',
                       txt=self.path / 'c2h4_pw_nosym.txt')
        pe.get_potential_energy()
        return pe.calc

    def c6h12_pw(self):
        pe = read(self['c2h4_pw_nosym'])
        pe = pe.repeat((3, 1, 1))
        pe.calc = GPAW(mode='pw', txt=self.path / 'c6h12_pw.txt')
        pe.get_potential_energy()
        return pe.calc

    def h2o_lcao(self):
        from ase.build import molecule
        atoms = molecule('H2O', cell=[8, 8, 8], pbc=1)
        atoms.center()
        atoms.calc = GPAW(mode='lcao', txt=self.path / 'h2o.txt')
        atoms.get_potential_energy()
        return atoms.calc

    def ti2o4(self, symmetry):
        pwcutoff = 400.0
        k = 4
        a = 4.59
        c = 2.96
        u = 0.305

        rutile_cell = [[a, 0, 0],
                       [0, a, 0],
                       [0, 0, c]]

        TiO2_basis = np.array([[0.0, 0.0, 0.0],
                               [0.5, 0.5, 0.5],
                               [u, u, 0.0],
                               [-u, -u, 0.0],
                               [0.5 + u, 0.5 - u, 0.5],
                               [0.5 - u, 0.5 + u, 0.5]])

        bulk_crystal = Atoms(symbols='Ti2O4',
                             scaled_positions=TiO2_basis,
                             cell=rutile_cell,
                             pbc=(1, 1, 1))

        tag = '_nosym' if symmetry == 'off' else ''
        bulk_calc = GPAW(mode=PW(pwcutoff),
                         nbands=42,
                         eigensolver=Davidson(1),
                         kpts={'size': (k, k, k), 'gamma': True},
                         xc='PBE',
                         occupations=FermiDirac(0.00001),
                         parallel={'band': 1},
                         symmetry=symmetry,
                         txt=self.path / f'ti2o4_pw{tag}.txt')

        bulk_crystal.calc = bulk_calc
        bulk_crystal.get_potential_energy()
        return bulk_calc

    def ti2o4_pw(self):
        return self.ti2o4({})

    def ti2o4_pw_nosym(self):
        return self.ti2o4('off')

    def si_pw(self):
        si = bulk('Si')
        calc = GPAW(mode='pw',
                    xc='LDA',
                    occupations=FermiDirac(width=0.001),
                    kpts={'size': (2, 2, 2), 'gamma': True},
                    txt=self.path / 'si_pw.txt')
        si.calc = calc
        si.get_potential_energy()
        return si.calc

    @with_band_cutoff(gpw='fancy_si_pw_wfs',
                      band_cutoff=8)  # 2 * (3s, 3p)
    def _fancy_si(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        xc = 'LDA'
        kpts = 4
        pw = 300
        occw = 0.01
        conv = {'bands': band_cutoff + 1,
                'density': 1.e-8}
        atoms = bulk('Si')
        atoms.center()

        tag = '_nosym' if symmetry == 'off' else ''
        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw),
            kpts={'size': (kpts, kpts, kpts), 'gamma': True},
            nbands=band_cutoff + 12,  # + 2 * (4s, 3d),
            occupations=FermiDirac(occw),
            convergence=conv,
            txt=self.path / f'fancy_si_pw{tag}.txt',
            symmetry=symmetry)

        atoms.get_potential_energy()
        return atoms.calc

    def fancy_si_pw(self):
        return self._fancy_si()

    def fancy_si_pw_nosym(self):
        return self._fancy_si(symmetry='off')

    def bn_pw(self):
        atoms = bulk('BN', 'zincblende', a=3.615)
        atoms.calc = GPAW(mode=PW(400),
                          kpts={'size': (2, 2, 2), 'gamma': True},
                          nbands=12,
                          convergence={'bands': 9},
                          occupations=FermiDirac(0.001),
                          txt=self.path / 'bn_pw.txt')
        atoms.get_potential_energy()
        return atoms.calc

    def hbn_pw(self):
        atoms = Graphene(symbol='B',
                         latticeconstant={'a': 2.5, 'c': 1.0},
                         size=(1, 1, 1))
        atoms[0].symbol = 'N'
        atoms.pbc = (1, 1, 0)
        atoms.center(axis=2, vacuum=3.0)
        atoms.calc = GPAW(mode=PW(400),
                          xc='LDA',
                          nbands=50,
                          occupations=FermiDirac(0.001),
                          parallel={'domain': 1},
                          convergence={'bands': 26},
                          kpts={'size': (3, 3, 1), 'gamma': True})
        atoms.get_potential_energy()
        return atoms.calc

    def graphene_pw(self):
        from ase.lattice.hexagonal import Graphene
        atoms = Graphene(symbol='C',
                         latticeconstant={'a': 2.45, 'c': 1.0},
                         size=(1, 1, 1))
        atoms.pbc = (1, 1, 0)
        atoms.center(axis=2, vacuum=4.0)
        ecut = 250
        nkpts = 6
        atoms.calc = GPAW(mode=PW(ecut),
                          kpts={'size': (nkpts, nkpts, 1), 'gamma': True},
                          nbands=len(atoms) * 6,
                          txt=self.path / 'graphene_pw.txt')
        atoms.get_potential_energy()
        return atoms.calc

    def mos2_pw(self):
        from ase.build import mx2
        atoms = mx2(formula='MoS2', kind='2H', a=3.184, thickness=3.127,
                    size=(1, 1, 1), vacuum=5)
        atoms.pbc = (1, 1, 1)
        ecut = 250
        nkpts = 6
        atoms.calc = GPAW(mode=PW(ecut),
                          xc='LDA',
                          kpts={'size': (nkpts, nkpts, 1), 'gamma': True},
                          occupations=FermiDirac(0.01),
                          txt=self.path / 'mos2_pw.txt')

        atoms.get_potential_energy()
        return atoms.calc

    def ni_pw_kpts333(self):
        from ase.dft.kpoints import monkhorst_pack
        # from gpaw.mpi import serial_comm
        Ni = bulk('Ni', 'fcc')
        Ni.set_initial_magnetic_moments([0.7])

        kpts = monkhorst_pack((3, 3, 3))

        calc = GPAW(mode='pw',
                    kpts=kpts,
                    occupations=FermiDirac(0.001),
                    setups={'Ni': '10'},
                    # communicator=serial_comm
                    )

        Ni.calc = calc
        Ni.get_potential_energy()
        calc.diagonalize_full_hamiltonian()
        # calc.write('Ni.gpw', mode='all')
        return calc

    def c_pw(self):
        atoms = bulk('C')
        atoms.center()
        calc = GPAW(mode=PW(150),
                    convergence={'bands': 6},
                    nbands=12,
                    kpts={'gamma': True, 'size': (2, 2, 2)},
                    xc='LDA')

        atoms.calc = calc
        atoms.get_potential_energy()
        return atoms.calc

    def nicl2_pw(self):
        from ase.build import mx2

        # Define input parameters
        xc = 'LDA'
        kpts = 6
        pw = 300
        occw = 0.01
        conv = {'density': 1.e-8,
                'forces': 1.e-8}

        a = 3.502
        thickness = 2.617
        vacuum = 3.0
        mm = 2.0

        # Set up atoms
        atoms = mx2(formula='NiCl2', kind='1T', a=a,
                    thickness=thickness, vacuum=vacuum)
        atoms.set_initial_magnetic_moments([mm, 0.0, 0.0])
        # Use pbc to allow for real-space density interpolation
        atoms.pbc = True

        # Set up calculator
        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw,
                    # Interpolate the density in real-space
                    interpolation=3),
            kpts={'size': (kpts, kpts, 1), 'gamma': True},
            occupations=FermiDirac(occw),
            convergence=conv,
            txt=self.path / 'nicl2_pw.txt')

        atoms.get_potential_energy()

        return atoms.calc

    @with_band_cutoff(gpw='v2br4_pw_wfs',
                      band_cutoff=28)  # V(4s,3d) = 6, Br(4s,4p) = 4
    def _v2br4(self, *, band_cutoff, symmetry=None):
        from ase.build import mx2

        if symmetry is None:
            symmetry = {}

        # Define input parameters
        xc = 'LDA'
        kpts = 4
        pw = 300
        occw = 0.01
        conv = {'density': 1.e-8,
                'bands': band_cutoff + 1}

        a = 3.840
        thickness = 2.897
        vacuum = 3.0
        mm = 3.0

        # Set up atoms
        atoms = mx2(formula='VBr2', kind='1T', a=a,
                    thickness=thickness, vacuum=vacuum)
        atoms = atoms.repeat((1, 2, 1))
        atoms.set_initial_magnetic_moments([mm, 0.0, 0.0, -mm, 0.0, 0.0])
        # Use pbc to allow for real-space density interpolation
        atoms.pbc = True

        # Set up calculator
        tag = '_nosym' if symmetry == 'off' else ''
        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw,
                    # Interpolate the density in real-space
                    interpolation=3),
            kpts={'size': (kpts, kpts // 2, 1), 'gamma': True},
            setups={'V': '5'},
            nbands=band_cutoff + 12,
            occupations=FermiDirac(occw),
            convergence=conv,
            symmetry=symmetry,
            txt=self.path / f'v2br4_pw{tag}.txt')

        atoms.get_potential_energy()

        return atoms.calc

    def v2br4_pw(self):
        return self._v2br4()

    def v2br4_pw_nosym(self):
        return self._v2br4(symmetry='off')

    @with_band_cutoff(gpw='fe_pw_wfs',
                      band_cutoff=9)  # 4s, 4p, 3d = 9
    def _fe(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        """See also the fe_fixture_test.py test."""
        xc = 'LDA'
        kpts = 4
        pw = 300
        occw = 0.01
        conv = {'bands': band_cutoff + 1,
                'density': 1.e-8}
        a = 2.867
        mm = 2.21
        atoms = bulk('Fe', 'bcc', a=a)
        atoms.set_initial_magnetic_moments([mm])
        atoms.center()
        tag = '_nosym' if symmetry == 'off' else ''

        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw),
            kpts={'size': (kpts, kpts, kpts)},
            nbands=band_cutoff + 9,
            occupations=FermiDirac(occw),
            convergence=conv,
            txt=self.path / f'fe_pw{tag}.txt',
            symmetry=symmetry)

        atoms.get_potential_energy()
        return atoms.calc

    def fe_pw(self):
        return self._fe()

    def fe_pw_nosym(self):
        return self._fe(symmetry='off')

    @with_band_cutoff(gpw='co_pw_wfs',
                      band_cutoff=14)  # 2 * (4s + 3d)
    def _co(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        # ---------- Inputs ---------- #

        # Atomic configuration
        a = 2.5071
        c = 4.0695
        mm = 1.6
        atoms = bulk('Co', 'hcp', a=a, c=c)
        atoms.set_initial_magnetic_moments([mm, mm])
        atoms.center()

        # Ground state parameters
        xc = 'LDA'
        occw = 0.01
        ebands = 2 * 2  # extra bands for ground state calculation
        pw = 200
        conv = {'density': 1e-8,
                'forces': 1e-8,
                'bands': band_cutoff + 1}

        # ---------- Calculation ---------- #

        tag = '_nosym' if symmetry == 'off' else ''
        atoms.calc = GPAW(xc=xc,
                          mode=PW(pw),
                          kpts={'size': (4, 4, 4), 'gamma': True},
                          occupations=FermiDirac(occw),
                          convergence=conv,
                          nbands=band_cutoff + ebands,
                          symmetry=symmetry,
                          txt=self.path / f'co_pw{tag}.txt')

        atoms.get_potential_energy()
        return atoms.calc

    def co_pw(self):
        return self._co()

    def co_pw_nosym(self):
        return self._co(symmetry='off')

    @with_band_cutoff(gpw='srvo3_pw_wfs',
                      band_cutoff=20)
    def _srvo3(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}

        nk = 3
        cell = bulk('V', 'sc', a=3.901).cell
        atoms = Atoms('SrVO3', cell=cell, pbc=True,
                      scaled_positions=((0.5, 0.5, 0.5),
                                        (0, 0, 0),
                                        (0, 0.5, 0),
                                        (0, 0, 0.5),
                                        (0.5, 0, 0)))
        # Ground state parameters
        xc = 'LDA'
        occw = 0.01
        ebands = 10  # extra bands for ground state calculation
        pw = 200
        conv = {'density': 1e-8,
                'bands': band_cutoff + 1}

        # ---------- Calculation ---------- #

        tag = '_nosym' if symmetry == 'off' else ''
        atoms.calc = GPAW(xc=xc,
                          mode=PW(pw),
                          kpts={'size': (nk, nk, nk), 'gamma': True},
                          occupations=FermiDirac(occw),
                          convergence=conv,
                          nbands=band_cutoff + ebands,
                          symmetry=symmetry,
                          txt=self.path / f'srvo3_pw{tag}.txt')

        atoms.get_potential_energy()
        return atoms.calc

    def srvo3_pw(self):
        return self._srvo3()

    def srvo3_pw_nosym(self):
        return self._srvo3(symmetry='off')

    @with_band_cutoff(gpw='al_pw_wfs',
                      band_cutoff=10)  # 3s, 3p, 4s, 3d
    def _al(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        xc = 'LDA'
        kpts = 4
        pw = 300
        occw = 0.01
        conv = {'bands': band_cutoff + 1,
                'density': 1.e-8}
        a = 4.043
        atoms = bulk('Al', 'fcc', a=a)
        atoms.center()
        tag = '_nosym' if symmetry == 'off' else ''

        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw),
            kpts={'size': (kpts, kpts, kpts), 'gamma': True},
            nbands=band_cutoff + 4,  # + 4p, 5s
            occupations=FermiDirac(occw),
            convergence=conv,
            txt=self.path / f'al_pw{tag}.txt',
            symmetry=symmetry)

        atoms.get_potential_energy()
        return atoms.calc

    def al_pw(self):
        return self._al()

    def al_pw_nosym(self):
        return self._al(symmetry='off')

    def ag_plusU_pw(self):
        xc = 'LDA'
        kpts = 2
        nbands = 6
        pw = 300
        occw = 0.01
        conv = {'bands': nbands,
                'density': 1e-12}
        a = 4.07
        atoms = bulk('Ag', 'fcc', a=a)
        atoms.center()

        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw),
            kpts={'size': (kpts, kpts, kpts), 'gamma': True},
            setups={'Ag': '11:d,2.0,0'},
            nbands=nbands,
            occupations=FermiDirac(occw),
            convergence=conv,
            parallel={'domain': 1},
            txt=self.path / 'ag_pw.txt')

        atoms.get_potential_energy()

        atoms.calc.diagonalize_full_hamiltonian()

        return atoms.calc

    def gaas_pw_nosym(self):
        return self._gaas(symmetry='off')

    def gaas_pw(self):
        return self._gaas()

    @with_band_cutoff(gpw='gaas_pw_wfs',
                      band_cutoff=8)
    def _gaas(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        nk = 4
        cell = bulk('Ga', 'fcc', a=5.68).cell
        atoms = Atoms('GaAs', cell=cell, pbc=True,
                      scaled_positions=((0, 0, 0), (0.25, 0.25, 0.25)))
        tag = '_nosym' if symmetry == 'off' else ''
        conv = {'bands': band_cutoff + 1,
                'density': 1.e-8}

        calc = GPAW(mode=PW(400),
                    xc='LDA',
                    occupations=FermiDirac(width=0.01),
                    convergence=conv,
                    nbands=band_cutoff + 1,
                    kpts={'size': (nk, nk, nk), 'gamma': True},
                    txt=self.path / f'gs_GaAs{tag}.txt',
                    symmetry=symmetry)

        atoms.calc = calc
        atoms.get_potential_energy()
        return atoms.calc

    def h_pw210_rmmdiis(self):
        return self._pw_210_rmmdiis(Atoms('H'), hund=True, nbands=4)

    def h2_pw210_rmmdiis(self):
        return self._pw_210_rmmdiis(Atoms('H2', [(0, 0, 0), (0, 0, 0.7413)]),
                                    nbands=8)

    def _pw_210_rmmdiis(self, atoms, **kwargs):
        atoms.set_pbc(True)
        atoms.set_cell((2., 2., 3.))
        atoms.center()
        calc = GPAW(mode=PW(210, force_complex_dtype=True),
                    eigensolver='rmm-diis',
                    xc='LDA',
                    basis='dzp',
                    parallel={'domain': 1},
                    convergence={'density': 1.e-6},
                    **kwargs)
        atoms.calc = calc
        atoms.get_potential_energy()
        calc.diagonalize_full_hamiltonian(nbands=80)
        return calc


class GPAWPlugin:
    def __init__(self):
        if world.rank == -1:
            print()
            info()

    def pytest_terminal_summary(self, terminalreporter, exitstatus, config):
        from gpaw.mpi import size
        terminalreporter.section('GPAW-MPI stuff')
        terminalreporter.write(f'size: {size}\n')


@pytest.fixture
def sg15_hydrogen():
    from io import StringIO
    from gpaw.test.pseudopotential.H_sg15 import pp_text
    from gpaw.upf import read_sg15
    # We can't easily load a non-python file from the test suite.
    # Therefore we load the pseudopotential from a Python file.
    return read_sg15(StringIO(pp_text))


def pytest_configure(config):
    # Allow for fake cupy:
    os.environ['GPAW_CPUPY'] = '1'

    if world.rank != 0:
        try:
            tw = config.get_terminal_writer()
        except AttributeError:
            pass
        else:
            tw._file = devnull
    config.pluginmanager.register(GPAWPlugin(), 'pytest_gpaw')
    for line in [
        'ci: test included in CI',
        'do: Direct optimization',
        'dscf: Delta-SCF',
        'elph: Electron-phonon',
        'fast: fast test',
        'gllb: GLLBSC tests',
        'gpu: GPU test',
        'hybrids: Hybrid functionals',
        'intel: fails on INTEL toolchain',
        'kspair: tests of kspair in the response code',
        'later: know failure for new refactored GPAW',
        'legacy: Old stuff that will be removed later',
        'libxc: LibXC requirered',
        'lrtddft: Linear-response TDDFT',
        'mgga: MGGA test',
        'mom: MOM',
        'ofdft: Orbital-free DFT',
        'response: tests of the response code',
        'rpa: tests of RPA',
        'rttddft: Real-time TDDFT',
        'serial: run in serial only',
        'slow: slow test',
        'soc: Spin-orbit coupling',
        'stress: Calculation of stress tensor',
        'wannier: Wannier functions']:
        config.addinivalue_line('markers', line)


def pytest_runtest_setup(item):
    """Skip some tests.

    If:

    * they depend on libxc and GPAW is not compiled with libxc
    * they are before $PYTEST_START_AFTER
    """
    from gpaw import libraries

    if world.size > 1:
        for mark in item.iter_markers():
            if mark.name == 'serial':
                pytest.skip('Only run in serial')

    if item.location[0] <= os.environ.get('PYTEST_START_AFTER', ''):
        pytest.skip('Not after $PYTEST_START_AFTER')
        return

    if libraries['libxc']:
        return

    if any(mark.name in {'libxc', 'mgga'}
           for mark in item.iter_markers()):
        pytest.skip('No LibXC.')


@pytest.fixture
def scalapack():
    """Skip if not compiled with sl.

    This fixture otherwise does not return or do anything."""
    from gpaw.utilities import compiled_with_sl
    if not compiled_with_sl():
        pytest.skip(reason='no scalapack')


@pytest.fixture
def needs_ase_master():
    from ase.utils.filecache import MultiFileJSONCache
    try:
        MultiFileJSONCache('bla-bla', comm=None)
    except TypeError:
        pytest.skip(reason='ASE is too old')


def pytest_report_header(config, startdir):
    # Use this to add custom information to the pytest printout.
    yield f'GPAW MPI rank={world.rank}, size={world.size}'



import numpy as np

from gpaw.new.eigensolver import Eigensolver
from gpaw.new.lcao.hamiltonian import HamiltonianMatrixCalculator
from gpaw.new.lcao.wave_functions import LCAOWaveFunctions


class LCAOEigensolver(Eigensolver):
    def __init__(self, basis):
        self.basis = basis

    def iterate(self, state, hamiltonian) -> float:
        matrix_calculator = hamiltonian.create_hamiltonian_matrix_calculator(
            state)

        for wfs in state.ibzwfs:
            self.iterate1(wfs, matrix_calculator)
        return 0.0

    def iterate1(self,
                 wfs: LCAOWaveFunctions,
                 matrix_calculator: HamiltonianMatrixCalculator):
        H_MM = matrix_calculator.calculate_matrix(wfs)
        eig_M = H_MM.eighg(wfs.L_MM, wfs.domain_comm)
        N = min(len(eig_M), wfs.nbands)
        wfs._eig_n = np.empty(wfs.nbands)
        wfs._eig_n[:N] = eig_M[:N]
        wfs.C_nM.data[:N] = H_MM.data.T[:N]

        # Make sure wfs.C_nM and (lazy) wfs.P_ani are in sync:
        wfs._P_ani = None



"""The purpose of this test is to make sure that wfs-dependent
functionals do not produce wrong results due to incorrect saving of
the wfs object when calc.set() is used.

It works by converging a calculation with (1, 2, 1) kpts, then setting
to (4, 1, 1) kpts and comparing.  That should yield the same result as
converging (4, 1, 1) right form the start.

This is just one example of the many inconsistencies that can be
caused by calls to calc.set().  That method should therefore only be
used with the utmost care."""

import pytest
from ase import Atoms
from gpaw import GPAW, Mixer
from gpaw.test import equal


@pytest.mark.libxc
@pytest.mark.legacy
def test_pathological_nonlocalset():
    atoms = Atoms('HF', positions=[(0.0, 0.0, 0.0), (1.0, 0.0, 0.0)])
    atoms.set_pbc((True, True, True))
    atoms.set_cell((2.01, 2.01, 2.01))  # make sure we get 12 gpts and not 8

    def MGGA_fail():
        calc = GPAW(xc='TPSS',
                    mixer=Mixer(0.5, 5, 50.0),
                    eigensolver='cg',
                    kpts=(1, 2, 1),
                    convergence={'density': 1e-8})
        atoms.calc = calc
        atoms.get_potential_energy()
        calc.set(kpts=(4, 1, 1))
        return atoms.get_potential_energy()

    def MGGA_work():
        calc = GPAW(xc='TPSS', mixer=Mixer(0.5, 5, 50.0),
                    eigensolver='cg', kpts=(4, 1, 1),
                    convergence={'density': 1e-8})

        atoms.calc = calc
        return atoms.get_potential_energy()

    def GLLBSC_fail():
        calc = GPAW(xc='GLLBSC', mixer=Mixer(0.5, 5, 50.0),
                    eigensolver='cg', kpts=(1, 2, 1),
                    convergence={'density': 1e-8})

        atoms.calc = calc
        atoms.get_potential_energy()
        calc.set(kpts=(4, 1, 1))
        return atoms.get_potential_energy()

    def GLLBSC_work():
        calc = GPAW(xc='GLLBSC', mixer=Mixer(0.5, 5, 50.0),
                    eigensolver='cg', kpts=(4, 1, 1),
                    convergence={'density': 1e-8})

        atoms.calc = calc
        return atoms.get_potential_energy()

    a = GLLBSC_fail()
    b = GLLBSC_work()
    equal(a, b, 1e-5)
    c = MGGA_fail()
    d = MGGA_work()
    equal(c, d, 1e-5)



import numpy as np
import pytest
from ase.build import bulk
from ase.parallel import parprint

from gpaw import GPAW, PW, FermiDirac


@pytest.mark.mgga
def test_pw_fe_stress_mgga(in_tmp_dir):
    xc = 'revTPSS'
    m = [2.9]
    fe = bulk('Fe')
    fe.set_initial_magnetic_moments(m)
    k = 3
    fe.calc = GPAW(mode=PW(800),
                   h=0.15,
                   occupations=FermiDirac(width=0.03),
                   xc=xc,
                   kpts=(k, k, k),
                   convergence={'energy': 1e-8},
                   parallel={'domain': 1, 'augment_grids': True},
                   txt='fe.txt')

    fe.set_cell(np.dot(fe.cell,
                       [[1.02, 0, 0.03],
                        [0, 0.99, -0.02],
                        [0.2, -0.01, 1.03]]),
                scale_atoms=True)

    fe.get_potential_energy()

    # Trigger nasty bug (fixed in !486):
    fe.calc.wfs.pt.blocksize = fe.calc.wfs.pd.maxmyng - 1

    s_analytical = fe.get_stress()
    # Calculated numerical stress once, store here to speed up test
    # numerical stresses:
    # revTPSS stress: [0.03113369 -0.05080607 -0.03739338
    #                  -0.03096389  0.21181234  0.0114693]
    s_numerical = np.array([0.03113369, -0.05080607, -0.03739338,
                            -0.03096389, 0.21181234, 0.0114693])
    # s_numerical = fe.calc.calculate_numerical_stress(fe, 1e-5)
    s_err = s_numerical - s_analytical

    parprint('Analytical stress:\n', s_analytical)
    parprint('Numerical stress:\n', s_numerical)
    parprint('Error in stress:\n', s_err)
    assert np.all(abs(s_err) < 1e-4)



import pytest
from gpaw.utilities import compiled_with_libvdwxc
from ase.build import molecule
from gpaw import GPAW, Mixer, Davidson, PW
from gpaw.xc.libvdwxc import vdw_df
from gpaw.mpi import world

pytestmark = pytest.mark.skipif(not compiled_with_libvdwxc(),
                                reason='not compiled_with_libvdwxc()')


@pytest.mark.later
def test_vdw_libvdwxc_h2(in_tmp_dir):
    system = molecule('H2')
    system.center(vacuum=1.0)
    system.pbc = 1

    def calculate(**kwargs1):
        kwargs = dict(mode=mode,
                      basis='sz(dzp)',
                      eigensolver=Davidson(4) if mode != 'lcao' else None,
                      xc=vdw_df(),
                      h=0.25,
                      convergence=dict(energy=1e-6),
                      mixer=Mixer(0.5, 5, 10.))
        kwargs.update(kwargs1)
        calc = GPAW(**kwargs)
        system.calc = calc
        system.get_potential_energy()
        return calc

    fddomainpar = min(2, world.size)

    for mode in ['fd', 'pw', 'lcao']:
        kwargs = {}
        if mode == 'pw':
            kwargs['mode'] = PW(250)
        else:
            kwargs['parallel'] = {'domain': fddomainpar}
        calc = calculate(**kwargs)

    E1 = calc.get_potential_energy()
    calc.write('dump.libvdwxc.gpw')
    _ = GPAW('dump.libvdwxc.gpw', txt='restart.txt',
             parallel={'domain': fddomainpar})
    system2 = calc.get_atoms()

    # Verify same energy after restart
    E2 = system2.get_potential_energy()
    assert abs(E2 - E1) < 1e-14  # Should be exact

    # Trigger recaclulation of essentially same system
    system2.positions[0, 0] += 1e-13
    print('reconverge')
    E3 = system2.get_potential_energy()
    err2 = abs(E3 - E2)
    print('error', err2)
    assert err2 < 1e-6  # Around SCF precision



from math import pi, sqrt

from ase.units import Hartree
import numpy as np
from numpy.fft import fftn

from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.pair_density import PairDensity2 as PairDensity
from gpaw.poisson import PoissonSolver, FFTPoissonSolver
from gpaw.utilities import unpack, packed_index, unpack2
from gpaw.utilities.ewald import madelung
from gpaw.utilities.tools import construct_reciprocal, tri2full, symmetrize
from gpaw.utilities.gauss import Gaussian
from gpaw.utilities.blas import r2k


def get_vxc(paw, spin=0, U=None):
    """Calculate matrix elements of the xc-potential."""
    assert not paw.hamiltonian.xc.xcfunc.orbital_dependent, "LDA/GGA's only"
    assert paw.wfs.dtype == float, 'Complex waves not implemented'

    if U is not None:  # Rotate xc matrix
        return np.dot(U.T.conj(), np.dot(get_vxc(paw, spin), U))

    gd = paw.hamiltonian.gd
    psit_nG = paw.wfs.kpt_u[spin].psit_nG[:]
    if paw.density.nt_sg is None:
        paw.density.interpolate_pseudo_density()
    nt_g = paw.density.nt_sg[spin]
    vxct_g = paw.density.finegd.zeros()
    paw.hamiltonian.xc.get_energy_and_potential(nt_g, vxct_g)
    vxct_G = gd.empty()
    paw.hamiltonian.restrict(vxct_g, vxct_G)
    Vxc_nn = np.zeros((paw.wfs.bd.nbands, paw.wfs.bd.nbands))

    # Apply pseudo part
    r2k(.5 * gd.dv, psit_nG, vxct_G * psit_nG, .0, Vxc_nn)  # lower triangle
    tri2full(Vxc_nn, 'L')  # Fill in upper triangle from lower
    gd.comm.sum(Vxc_nn)

    # Add atomic PAW corrections
    for a, P_ni in paw.wfs.kpt_u[spin].P_ani.items():
        D_sp = paw.density.D_asp[a][:]
        H_sp = np.zeros_like(D_sp)
        paw.wfs.setups[a].xc_correction.calculate_energy_and_derivatives(
            D_sp, H_sp)
        H_ii = unpack(H_sp[spin])
        Vxc_nn += np.dot(P_ni, np.dot(H_ii, P_ni.T))
    return Vxc_nn * Hartree


class Coulomb:
    """Class used to evaluate two index coulomb integrals."""
    def __init__(self, gd, poisson=None):
        """Class should be initialized with a grid_descriptor 'gd' from
           the gpaw module.
        """
        self.gd = gd
        self.poisson = poisson

    def load(self, method):
        """Make sure all necessary attributes have been initialized"""

        assert method in ('real', 'recip_gauss', 'recip_ewald'),\
            str(method) + ' is an invalid method name,\n' +\
            'use either real, recip_gauss, or recip_ewald'

        if method.startswith('recip'):
            if self.gd.comm.size > 1:
                raise RuntimeError("Cannot do parallel FFT, use method='real'")
            if not hasattr(self, 'k2'):
                self.k2, self.N3 = construct_reciprocal(self.gd)
            if method.endswith('ewald') and not hasattr(self, 'ewald'):
                # cutoff radius
                assert self.gd.orthogonal
                rc = 0.5 * np.average(self.gd.cell_cv.diagonal())
                # ewald potential: 1 - cos(k rc)
                self.ewald = (np.ones(self.gd.n_c) -
                              np.cos(np.sqrt(self.k2) * rc))
                # lim k -> 0 ewald / k2
                self.ewald[0, 0, 0] = 0.5 * rc**2
            elif method.endswith('gauss') and not hasattr(self, 'ng'):
                gauss = Gaussian(self.gd)
                self.ng = gauss.get_gauss(0) / sqrt(4 * pi)
                self.vg = gauss.get_gauss_pot(0) / sqrt(4 * pi)
        else:  # method == 'real'
            if not hasattr(self, 'solve'):
                if self.poisson is not None:
                    self.solve = self.poisson.solve
                else:
                    solver = PoissonSolver('fd', nn=2, eps=1e-12)
                    solver.set_grid_descriptor(self.gd)
                    # solver.initialize()
                    self.solve = solver.solve

    def coulomb(self, n1, n2=None, Z1=None, Z2=None, method='recip_gauss'):
        """Evaluates the coulomb integral of n1 and n2

        The coulomb integral is defined by::

                                      *
                      /    /      n1(r)  n2(r')
          (n1 | n2) = | dr | dr'  -------------,
                      /    /         |r - r'|

        where n1 and n2 could be complex.

        real:
           Evaluate directly in real space using gaussians to neutralize
           density n2, such that the potential can be generated by standard
           procedures

        recip_ewald:
           Evaluate by Fourier transform.
           Divergence at division by k^2 is avoided by utilizing the Ewald /
           Tuckermann trick, which formally requires the densities to be
           localized within half of the unit cell.

        recip_gauss:
           Evaluate by Fourier transform.
           Divergence at division by k^2 is avoided by removing total charge
           of n1 and n2 with gaussian density ng::

                                                   *          *    *
            (n1|n2) = (n1 - Z1 ng|n2 - Z2 ng) + (Z2 n1 + Z1 n2 - Z1 Z2 ng | ng)

           The evaluation of the integral (n1 - Z1 ng|n2 - Z2 ng) is done in
           k-space using FFT techniques.
        """
        self.load(method)
        # determine integrand using specified method
        if method == 'real':
            I = self.gd.zeros()
            if n2 is None:
                n2 = n1
                Z2 = Z1
            self.solve(I, n2, charge=Z2, zero_initial_phi=True)
            I += madelung(self.gd.cell_cv) * self.gd.integrate(n2)
            I *= n1.conj()
        elif method == 'recip_ewald':
            n1k = fftn(n1)
            if n2 is None:
                n2k = n1k
            else:
                n2k = fftn(n2)
            I = n1k.conj() * n2k * self.ewald * 4 * pi / (self.k2 * self.N3)
        else:  # method == 'recip_gauss':
            # Determine total charges
            if Z1 is None:
                Z1 = self.gd.integrate(n1)
            if Z2 is None and n2 is not None:
                Z2 = self.gd.integrate(n2)

            # Determine the integrand of the neutral system
            # (n1 - Z1 ng)* int dr'  (n2 - Z2 ng) / |r - r'|
            nk1 = fftn(n1 - Z1 * self.ng)
            if n2 is None:
                I = abs(nk1)**2 * 4 * pi / (self.k2 * self.N3)
            else:
                nk2 = fftn(n2 - Z2 * self.ng)
                I = nk1.conj() * nk2 * 4 * pi / (self.k2 * self.N3)

            # add the corrections to the integrand due to neutralization
            if n2 is None:
                I += (2 * np.real(np.conj(Z1) * n1) -
                      abs(Z1)**2 * self.ng) * self.vg
            else:
                I += (np.conj(Z1) * n2 + Z2 * n1.conj() -
                      np.conj(Z1) * Z2 * self.ng) * self.vg
        if n1.dtype == float and (n2 is None or n2.dtype == float):
            return np.real(self.gd.integrate(I))
        else:
            return self.gd.integrate(I)


class CoulombNEW:
    def __init__(self, gd, setups, spos_ac, fft=False):
        assert gd.comm.size == 1
        self.rhot1_G = gd.empty()
        self.rhot2_G = gd.empty()
        self.pot_G = gd.empty()
        self.dv = gd.dv
        if fft:
            self.poisson = FFTPoissonSolver()
        else:
            self.poisson = PoissonSolver(name='fd', nn=3, eps=1e-12)
        self.poisson.set_grid_descriptor(gd)
        self.setups = setups

        # Set coarse ghat
        self.Ghat = LFC(gd, [setup.ghat_l for setup in setups],
                        integral=sqrt(4 * pi))
        self.Ghat.set_positions(spos_ac)

    def calculate(self, nt1_G, nt2_G, P1_ap, P2_ap):
        I = 0.0
        self.rhot1_G[:] = nt1_G
        self.rhot2_G[:] = nt2_G

        Q1_aL = {}
        Q2_aL = {}
        for a, P1_p in P1_ap.items():
            P2_p = P2_ap[a]
            setup = self.setups[a]

            # Add atomic corrections to integral
            I += 2 * np.dot(P1_p, np.dot(setup.M_pp, P2_p))

            # Add compensation charges to pseudo densities
            Q1_aL[a] = np.dot(P1_p, setup.Delta_pL)
            Q2_aL[a] = np.dot(P2_p, setup.Delta_pL)
        self.Ghat.add(self.rhot1_G, Q1_aL)
        self.Ghat.add(self.rhot2_G, Q2_aL)

        # Add coulomb energy of compensated pseudo densities to integral
        self.poisson.solve(self.pot_G, self.rhot2_G, charge=None,
                           zero_initial_phi=True)
        I += np.vdot(self.rhot1_G, self.pot_G) * self.dv

        return I * Hartree


class HF:
    def __init__(self, paw):
        paw.initialize_positions()
        self.nspins = paw.wfs.nspins
        self.nbands = paw.wfs.bd.nbands
        self.restrict = paw.hamiltonian.restrict
        self.pair_density = PairDensity(paw.density, paw.spos_ac,
                                        finegrid=True)
        self.dv = paw.wfs.gd.dv
        self.dtype = paw.wfs.dtype
        self.setups = paw.wfs.setups

        # Allocate space for matrices
        self.nt_G = paw.wfs.gd.empty()
        self.rhot_g = paw.density.finegd.empty()
        self.vt_G = paw.wfs.gd.empty()
        self.vt_g = paw.density.finegd.empty()
        self.poisson_solve = paw.hamiltonian.poisson.solve

    def apply(self, paw, u=0):
        H_nn = np.zeros((self.nbands, self.nbands), self.dtype)
        self.soft_pseudo(paw, H_nn, u=u)
        self.atomic_val_val(paw, H_nn, u=u)
        self.atomic_val_core(paw, H_nn, u=u)
        return H_nn * Hartree

    def soft_pseudo(self, paw, H_nn, h_nn=None, u=0):
        if h_nn is None:
            h_nn = H_nn
        kpt = paw.wfs.kpt_u[u]
        pd = self.pair_density
        deg = 2 / self.nspins
        fmin = 1e-9
        Htpsit_nG = np.zeros(kpt.psit_nG.shape, self.dtype)

        for n1 in range(self.nbands):
            psit1_G = kpt.psit_nG[n1]
            f1 = kpt.f_n[n1] / deg
            for n2 in range(n1, self.nbands):
                psit2_G = kpt.psit_nG[n2]
                f2 = kpt.f_n[n2] / deg
                if f1 < fmin and f2 < fmin:
                    continue

                pd.initialize(kpt, n1, n2)
                pd.get_coarse(self.nt_G)
                pd.add_compensation_charges(self.nt_G, self.rhot_g)
                self.poisson_solve(self.vt_g, -self.rhot_g,
                                   charge=-float(n1 == n2),
                                   zero_initial_phi=True)
                self.restrict(self.vt_g, self.vt_G)
                Htpsit_nG[n1] += f2 * self.vt_G * psit2_G
                if n1 != n2:
                    Htpsit_nG[n2] += f1 * self.vt_G * psit1_G

                v_aL = paw.density.ghat.dict()
                paw.density.ghat.integrate(self.vt_g, v_aL)
                for a, v_L in v_aL.items():
                    v_ii = unpack(np.dot(paw.wfs.setups[a].Delta_pL, v_L))
                    P_ni = kpt.P_ani[a]
                    h_nn[:, n1] += f2 * np.dot(P_ni, np.dot(v_ii, P_ni[n2]))
                    if n1 != n2:
                        h_nn[:, n2] += f1 * np.dot(P_ni,
                                                   np.dot(v_ii, P_ni[n1]))

        symmetrize(h_nn)  # Grrrr why!!! XXX

        # Fill in lower triangle
        r2k(0.5 * self.dv, kpt.psit_nG[:], Htpsit_nG, 1.0, H_nn)

        # Fill in upper triangle from lower
        tri2full(H_nn, 'L')

    def atomic_val_val(self, paw, H_nn, u=0):
        deg = 2 / self.nspins
        kpt = paw.wfs.kpt_u[u]
        for a, P_ni in kpt.P_ani.items():
            # Add atomic corrections to the valence-valence exchange energy
            # --
            # >  D   C     D
            # --  ii  iiii  ii
            setup = paw.wfs.setups[a]
            D_p = paw.density.D_asp[a][kpt.s]
            H_p = np.zeros_like(D_p)
            D_ii = unpack2(D_p)
            ni = len(D_ii)
            for i1 in range(ni):
                for i2 in range(ni):
                    A = 0.0
                    for i3 in range(ni):
                        p13 = packed_index(i1, i3, ni)
                        for i4 in range(ni):
                            p24 = packed_index(i2, i4, ni)
                            A += setup.M_pp[p13, p24] * D_ii[i3, i4]
                    p12 = packed_index(i1, i2, ni)
                    H_p[p12] -= 2 / deg * A / ((i1 != i2) + 1)
            H_nn += np.dot(P_ni, np.inner(unpack(H_p), P_ni.conj()))

    def atomic_val_core(self, paw, H_nn, u=0):
        kpt = paw.wfs.kpt_u[u]
        for a, P_ni in kpt.P_ani.items():
            dH_ii = unpack(-paw.wfs.setups[a].X_p)
            H_nn += np.dot(P_ni, np.inner(dH_ii, P_ni.conj()))




class Perturbation:
    """Abstract base class for various perturbation types.

    Specific perturbations must derive from this class. As a minimum a
    perturbation should provide an ``apply`` member function that implements
    the action of the perturbing potential onto a wave function.

    """
    
    def get_phase_cd(self):
        """Return phase for instances of class ``Transformer``.

        Only relevant for q-specific perturbations.

        """
        
        return None

    def has_q(self):
        """Is perturbation q-dependent."""

        return False
        
    def apply(self, psi_nG, y_nG, k, kplusq=None):
        """Apply perturbing potential to unperturbed wave-functions."""
        
        raise NotImplementedError('Implement in derived classes')



from ase import Atoms
from gpaw import GPAW


def test_ae_potential():
    h = Atoms('H',
              cell=[2.01, 2.01, 2.01],
              pbc=True)
    h.calc = GPAW(
        mode='lcao',
        setups='ae',
        xc='PBE')
    f = h.get_forces()
    assert abs(f).max() < 1e-14



# General modules
import pytest

from itertools import product

import numpy as np

# import matplotlib.pyplot as plt

# Script modules
from ase.units import Hartree

from gpaw import GPAW
import gpaw.mpi as mpi
from gpaw.response import ResponseGroundStateAdapter
from gpaw.response.frequencies import ComplexFrequencyDescriptor
from gpaw.response.jdos import JDOSCalculator
from gpaw.response.symmetry import KPointFinder
from gpaw.test.response.test_chiks import (generate_system_s,
                                           generate_qrel_q, get_q_c)
from gpaw.test.conftest import response_band_cutoff


@pytest.mark.response
@pytest.mark.kspair
@pytest.mark.parametrize('system,qrel',
                         product(generate_system_s(), generate_qrel_q()))
def test_jdos(in_tmp_dir, gpw_files, system, qrel):
    # ---------- Inputs ---------- #

    # What material, spin-component and q-vector to calculate the jdos for
    wfs, spincomponent = system
    q_c = get_q_c(wfs, qrel)

    # Where to evaluate the jdos
    omega_w = np.linspace(-10.0, 10.0, 321)
    eta = 0.2
    zd = ComplexFrequencyDescriptor.from_array(omega_w + 1.j * eta)

    # Calculation parameters (which should not affect the result)
    disable_syms_s = [True, False]
    bandsummation_b = ['double', 'pairwise']

    # ---------- Script ---------- #

    # Set up the ground state adapter based on the fixture
    calc = GPAW(gpw_files[wfs], parallel=dict(domain=1))
    nbands = response_band_cutoff[wfs]
    gs = ResponseGroundStateAdapter(calc)

    # Calculate the jdos manually
    serial_calc = GPAW(gpw_files[wfs], communicator=mpi.serial_comm)
    jdos_refcalc = MyManualJDOS(serial_calc)
    jdosref_w = jdos_refcalc.calculate(spincomponent, q_c,
                                       omega_w,
                                       eta=eta,
                                       nbands=nbands)

    # Calculate the jdos using the PairFunctionIntegrator module
    for disable_syms in disable_syms_s:
        for bandsummation in bandsummation_b:
            jdos_calc = JDOSCalculator(gs,
                                       nbands=nbands,
                                       disable_time_reversal=disable_syms,
                                       disable_point_group=disable_syms,
                                       bandsummation=bandsummation)
            jdos = jdos_calc.calculate(spincomponent, q_c, zd)
            jdos_w = jdos.array
            assert jdos_w == pytest.approx(jdosref_w)

        # plt.subplot()
        # plt.plot(wd.omega_w * Hartree, jdos_w)
        # plt.plot(wd.omega_w * Hartree, jdosref_w)
        # plt.title(f'{q_c} {spincomponent}')
        # plt.show()


class MyManualJDOS:
    def __init__(self, calc):
        self.calc = calc
        self.nspins = calc.wfs.nspins

        kd = calc.wfs.kd
        gd = calc.wfs.gd
        self.kd = kd
        self.kptfinder = KPointFinder(kd.bzk_kc)
        self.kweight = 1 / (gd.volume * len(kd.bzk_kc))

    def calculate(self, spincomponent, q_c, omega_w,
                  eta=0.2,
                  nbands=None):
        r"""Calculate the joint density of states:
                       __  __
                    1  \   \
        g_j(q, ω) = ‾  /   /  (f_nks - f_mk+qs') δ(ω-[ε_mk+qs' - ε_nks])
                    V  ‾‾  ‾‾
                       k   n,m

        for a given spin component specifying the spin transitions s -> s'.
        """
        q_c = np.asarray(q_c)
        # Internal frequencies in Hartree
        omega_w = omega_w / Hartree
        eta = eta / Hartree
        # Allocate array
        jdos_w = np.zeros_like(omega_w)
        
        for K1, k1_c in enumerate(self.kd.bzk_kc):
            # de = e2 - e1, df = f2 - f1
            de_t, df_t = self.get_transitions(K1, k1_c, q_c,
                                              spincomponent, nbands)

            if self.nspins == 1:
                df_t *= 2

            # Set up jdos
            delta_wt = self.delta(omega_w, eta, de_t)
            jdos_wt = - df_t[np.newaxis] * delta_wt

            # Sum over transitions
            jdos_w += np.sum(jdos_wt, axis=1)

        return self.kweight * jdos_w

    @staticmethod
    def delta(omega_w, eta, de_t):
        r"""Create lorentzian delta-functions

                ~ 1       η
        δ(ω-Δε) = ‾ ‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                  π (ω-Δε)^2 + η^2
        """
        x_wt = omega_w[:, np.newaxis] - de_t[np.newaxis]
        return eta / np.pi / (x_wt**2. + eta**2.)

    def get_transitions(self, K1, k1_c, q_c, spincomponent, nbands):
        assert isinstance(nbands, int)
        if spincomponent == '00':
            if self.nspins == 1:
                s1_s = [0]
                s2_s = [0]
            else:
                s1_s = [0, 1]
                s2_s = [0, 1]
        elif spincomponent == '+-':
            s1_s = [0]
            s2_s = [1]
        else:
            raise ValueError(spincomponent)

        # Find k_c + q_c
        K2 = self.kptfinder.find(k1_c + q_c)

        de_t = []
        df_t = []
        kd = self.kd
        calc = self.calc
        for s1, s2 in zip(s1_s, s2_s):
            # Get composite u=(s,k) indices and KPoint objects
            u1 = kd.bz2ibz_k[K1] * self.nspins + s1
            u2 = kd.bz2ibz_k[K2] * self.nspins + s2
            kpt1, kpt2 = calc.wfs.kpt_u[u1], calc.wfs.kpt_u[u2]

            # Extract eigenenergies and occupation numbers
            eps1_n = kpt1.eps_n[:nbands]
            eps2_n = kpt2.eps_n[:nbands]
            f1_n = kpt1.f_n[:nbands] / kpt1.weight
            f2_n = kpt2.f_n[:nbands] / kpt2.weight

            # Append data
            de_nm = eps2_n[:, np.newaxis] - eps1_n[np.newaxis]
            df_nm = f2_n[:, np.newaxis] - f1_n[np.newaxis]
            de_t += list(de_nm.flatten())
            df_t += list(df_nm.flatten())
        de_t = np.array(de_t)
        df_t = np.array(df_t)

        return de_t, df_t
        



from math import sqrt, pi

import pytest
import numpy as np
from ase import Atoms
from ase.units import Bohr, Hartree

from gpaw.mpi import rank, size
from gpaw import GPAW
from gpaw.external import ConstantElectricField
from gpaw.utilities import packed_index
from gpaw.pair_density import PairDensity

# Three ways to compute the polarizability of hydrogen:
# 1. Perturbation theory
# 2. Constant electric field
# 3. Middle of an electric dipole

# Note: The analytical value for the polarizability
# is 4.5 a0**3 (e.g. PRA 33, 3671), while the experimental
# value is 4.6 a0**3 (e.g. PR 133, A629).


@pytest.mark.skip(reason='TODO')
def test_stark_shift():
    from gpaw.point_charges import PointCharges
    to_au = Hartree / Bohr**2
    to_eVA = Hartree / Bohr

    def dipole_op(c, state1, state2, k=0, s=0):
        # Taken from KSSingle, maybe make this accessible in
        # KSSingle?
        wfs = c.wfs
        gd = wfs.gd

        kpt = None
        for i in wfs.kpt_u:
            if i.k == k and i.s == s:
                kpt = i

        pd = PairDensity(c)
        pd.initialize(kpt, state1, state2)

        # coarse grid contribution
        # <i|r|j> is the negative of the dipole moment (because of negative
        # e- charge)
        me = -gd.calculate_dipole_moment(pd.get())

        # augmentation contributions
        ma = np.zeros(me.shape)
        pos_av = c.get_atoms().get_positions() / Bohr
        for a, P_ni in kpt.P_ani.items():
            Ra = pos_av[a]
            Pi_i = P_ni[state1]
            Pj_i = P_ni[state2]
            Delta_pL = wfs.setups[a].Delta_pL
            ni = len(Pi_i)

            ma0 = 0
            ma1 = np.zeros(me.shape)
            for i in range(ni):
                for j in range(ni):
                    pij = Pi_i[i] * Pj_i[j]
                    ij = packed_index(i, j, ni)
                    # L=0 term
                    ma0 += Delta_pL[ij, 0] * pij
                    # L=1 terms
                    if wfs.setups[a].lmax >= 1:
                        # see spherical_harmonics.py for
                        # L=1:y L=2:z; L=3:x
                        ma1 += np.array([
                            Delta_pL[ij, 3], Delta_pL[ij, 1], Delta_pL[ij, 2]
                        ]) * pij
            ma += sqrt(4 * pi / 3) * ma1 + Ra * sqrt(4 * pi) * ma0
        gd.comm.sum(ma)

        me += ma

        return me * Bohr

    # Currently only works on a single processor
    assert size == 1

    maxfield = 0.01
    nfs = 5  # number of field
    nbands = 30  # number of bands
    h = 0.20  # grid spacing

    debug = not False

    if debug:
        txt = 'gpaw.out'
    else:
        txt = None

    test1 = True
    test2 = True
    test3 = True

    a0 = 6.0
    a = Atoms('H', positions=[[a0 / 2, a0 / 2, a0 / 2]], cell=[a0, a0, a0])

    alpha1 = None
    alpha2 = None
    alpha3 = None

    # Test 1

    if test1:
        c = GPAW(h=h,
                 nbands=nbands + 10,
                 spinpol=True,
                 hund=True,
                 xc='LDA',
                 eigensolver='cg',
                 convergence={
                     'bands': nbands,
                     'eigenstates': 3.3e-4
                 },
                 maxiter=1000,
                 txt=txt)
        a.calc = c
        a.get_potential_energy()

        o1 = c.get_occupation_numbers(spin=0)

        if o1[0] > 0.0:
            spin = 0
        else:
            spin = 1

        alpha = 0.0

        ev = c.get_eigenvalues(0, spin)
        for i in range(1, nbands):
            mu_x, mu_y, mu_z = dipole_op(c, 0, i, k=0, s=spin)

            alpha += mu_z**2 / (ev[i] - ev[0])

        alpha *= 2

        if rank == 0 and debug:
            print('From perturbation theory:')
            print('  alpha = ', alpha, ' A**2/eV')
            print('  alpha = ', alpha * to_au, ' Bohr**3')

        alpha1 = alpha

    ###

    c = GPAW(
        h=h,
        nbands=2,
        spinpol=True,
        hund=True,
        xc='LDA',
        # eigensolver  = 'cg',
        convergence={
            'bands': nbands,
            'eigenstates': 3.3e-4
        },
        maxiter=1000,
        txt=txt)
    a.calc = c

    # Test 2

    if test2:
        e = []
        e1s = []
        d = []
        fields = np.linspace(-maxfield, maxfield, nfs)
        for field in fields:
            if rank == 0 and debug:
                print(field)
            c = c.new(external=ConstantElectricField(field))
            a.calc = c
            etot = a.get_potential_energy()
            e += [etot]
            ev0 = c.get_eigenvalues(0)
            ev1 = c.get_eigenvalues(1)
            e1s += [min(ev0[0], ev1[0])]
            dip = c.get_dipole_moment()
            d += [dip[2]]

        pol1, dummy = np.polyfit(fields, d, 1)
        pol2, dummy1, dummy2 = np.polyfit(fields, e1s, 2)

        if rank == 0 and debug:
            print('From shift in 1s-state at constant electric field:')
            print('  alpha = ', -pol2, ' A**2/eV')
            print('  alpha = ', -pol2 * to_au, ' Bohr**3')

            print('From dipole moment at constant electric field:')
            print('  alpha = ', pol1, ' A**2/eV')
            print('  alpha = ', pol1 * to_au, ' Bohr**3')

            np.savetxt('ecf.out', np.transpose([fields, e, e1s, d]))

        assert abs(pol1 + pol2) < 0.0001

        alpha2 = (pol1 - pol2) / 2

    # Test 3

    if test3:
        pcd = 1000.0  # distance of the two point charges
        maxcharge = 100.0  # maximum charge on the point charge

        e = []
        e1s = []
        d = []
        charges = np.linspace(-maxcharge, maxcharge, nfs)
        fields = []
        for charge in charges:
            ex = PointCharges(positions=[[a0 / 2, a0 / 2, -pcd / 2 + a0 / 2],
                                         [a0 / 2, a0 / 2, pcd / 2 + a0 / 2]],
                              charges=[charge, -charge])
            c = c.new(external=ex)
            a.calc = c
            etot = a.get_potential_energy()
            e += [etot]
            ev0 = c.get_eigenvalues(0)
            ev1 = c.get_eigenvalues(1)
            e1s += [min(ev0[0], ev1[0])]
            dip = c.get_dipole_moment()
            d += [dip[2]]
            field = ex.get_taylor(position=a[0].position)[1][1]
            if rank == 0 and debug:
                print(field * to_eVA,
                      2 * charge / ((pcd / 2)**2) * Hartree * Bohr)
            fields += [field * to_eVA]

        pol1, dummy = np.polyfit(fields, d, 1)
        pol2, dummy1, dummy2 = np.polyfit(fields, e1s, 2)

        if rank == 0 and debug:
            print('From shift in 1s-state between two point charges:')
            print('  alpha = ', -pol2, ' A**2/eV')
            print('  alpha = ', -pol2 * to_au, ' Bohr**3')

            # print 'From dipole moment between two point charges:'
            # print '  alpha = ', pol1, ' A**2/eV'
            # print '  alpha = ', pol1*to_au, ' Bohr**3'

            np.savetxt('epc.out', np.transpose([fields, e, e1s, d]))

        alpha3 = alpha

    # This is a very, very rough test
    assert abs(alpha1 - alpha2) < 0.01
    assert abs(alpha3 - alpha2) < 0.01



from ase import Atoms
from ase.units import Hartree
from gpaw import GPAW, PoissonSolver, FermiDirac, Davidson, MixerSum
from gpaw.test import equal


def test_xc_revPBE_Li():
    a = 5.0
    n = 24
    li = Atoms('Li', magmoms=[1.0], cell=(a, a, a), pbc=True)

    params = dict(
        gpts=(n, n, n),
        nbands=1,
        xc=dict(name='oldPBE', stencil=1),
        poissonsolver=PoissonSolver(),
        mixer=MixerSum(0.6, 5, 10.0),
        eigensolver=Davidson(4),
        convergence=dict(eigenstates=4.5e-8),
        occupations=FermiDirac(0.0))

    li.calc = GPAW(**params)
    e = li.get_potential_energy() + li.calc.get_reference_energy()
    equal(e, -7.462 * Hartree, 1.4)

    params['xc'] = dict(name='revPBE', stencil=1)
    li.calc = GPAW(**params)
    erev = li.get_potential_energy() + li.calc.get_reference_energy()

    equal(erev, -7.487 * Hartree, 1.3)
    equal(e - erev, 0.025 * Hartree, 0.002 * Hartree)

    print(e, erev)
    energy_tolerance = 0.002
    equal(e, -204.381098849, energy_tolerance)  # svnversion 5252
    equal(erev, -205.012303379, energy_tolerance)  # svnversion 5252



import os

import pytest
from ase import Atoms

from gpaw import GPAW


@pytest.mark.ci
@pytest.mark.mgga
def test_fixdensity(in_tmp_dir):
    a = 2.5
    slab = Atoms('Li', cell=(a, a, 2 * a), pbc=1)
    slab.calc = GPAW(xc='revTPSS',
                     h=0.15,
                     kpts=(3, 3, 1), txt='li-1.txt',
                     parallel=dict(kpt=1))
    slab.get_potential_energy()
    slab.calc.write('li.gpw', mode='all')
    slab.calc.write('li_nowf.gpw')

    # Gamma point:
    e1 = slab.calc.get_eigenvalues(kpt=0)[0]
    f1 = slab.calc.get_fermi_level()

    kpts = [(0, 0, 0)]

    # Fix density and continue:
    calc = slab.calc.fixed_density(
        txt='li-2.txt',
        nbands=5,
        kpts=kpts)
    e2 = calc.get_eigenvalues(kpt=0)[0]
    f2 = calc.get_fermi_level()

    # Start from gpw-file:
    calc = GPAW('li.gpw', txt=None)
    calc = calc.fixed_density(
        txt='li-3.txt',
        nbands=5,
        kpts=kpts)
    e3 = calc.get_eigenvalues(kpt=0)[0]
    f3 = calc.get_fermi_level()

    assert f2 == pytest.approx(f1, abs=1e-10)
    assert f3 == pytest.approx(f1, abs=1e-10)
    assert e2 == pytest.approx(e1, abs=3e-5)
    assert e3 == pytest.approx(e1, abs=3e-5)

    if os.environ.get('GPAW_NEW'):
        return

    try:
        calc = GPAW('li.gpw',
                    txt='li-4.txt',
                    fixdensity=True,
                    nbands=5,
                    kpts=kpts,
                    symmetry='off')

        with pytest.warns(DeprecationWarning):
            calc.get_potential_energy()
        calc.get_eigenvalues(kpt=0)[0]
    except ValueError:
        pass
    else:
        raise RuntimeError

    try:
        calc = GPAW('li_nowf.gpw', txt=None)
        calc = calc.fixed_density(
            txt='li-3.txt',
            nbands=5,
            kpts=kpts)
    except ValueError:
        pass
    else:
        raise RuntimeError



"""Zero-field splitting.

See::

    Spin decontamination for magnetic dipolar coupling calculations:
    Application to high-spin molecules and solid-state spin qubits

    Timur Biktagirov, Wolf Gero Schmidt, and Uwe Gerstmann

    Phys. Rev. Research 2, 022024(R) – Published 30 April 2020

"""
from math import pi
from typing import List, Tuple, Dict

import numpy as np
from ase.units import Bohr, Ha, _c, _e, _hplanck

from gpaw.calculator import GPAW
from gpaw.grid_descriptor import GridDescriptor
from gpaw.typing import Array1D, Array2D, Array4D
from gpaw.hyperfine import alpha  # fine-structure constant: ~ 1 / 137
from gpaw.setup import Setup
from gpaw.pw.lfc import PWLFC
from gpaw.pw.descriptor import PWDescriptor
from gpaw.mpi import serial_comm


def zfs(calc: GPAW,
        method: int = 1) -> Array2D:
    """Zero-field splitting.

    Calculate magnetic dipole coupling tensor in eV.
    """
    (kpt1, kpt2), = calc.wfs.kpt_qs  # spin-polarized and gamma only

    nocc1 = (kpt1.f_n > 0.5).sum()
    nocc2 = (kpt2.f_n > 0.5).sum()

    assert nocc1 == nocc2 + 2, (nocc1, nocc2)

    if method == 1:
        wf1 = WaveFunctions.from_calc(calc, 0, nocc1 - 2, nocc1)
        wf12 = [wf1]
    else:
        wf1 = WaveFunctions.from_calc(calc, 0, 0, nocc1)
        wf2 = WaveFunctions.from_calc(calc, 1, 0, nocc2)
        wf12 = [wf1, wf2]

    D_vv = np.zeros((3, 3))

    if calc.world.rank == 0:
        compensation_charge = create_compensation_charge(wf1.setups,
                                                         wf1.pd,
                                                         calc.spos_ac)
        for wfa in wf12:
            for wfb in wf12:
                d_vv = zfs1(wfa, wfb, compensation_charge)
                D_vv += d_vv

    calc.world.broadcast(D_vv, 0)

    return D_vv


class WaveFunctions:
    def __init__(self,
                 psit_nR: Array4D,
                 P_ani: Dict[int, Array2D],
                 spin: int,
                 setups: List[Setup],
                 gd: GridDescriptor = None,
                 pd: PWDescriptor = None):
        """Container for wave function in real-space and projections."""

        self.pd = pd or PWDescriptor(ecut=None, gd=gd)
        self.psit_nR = psit_nR
        self.P_ani = P_ani
        self.spin = spin
        self.setups = setups

    @staticmethod
    def from_calc(calc: GPAW, spin: int, n1: int, n2: int) -> 'WaveFunctions':
        """Create WaveFunctions object GPAW calculation."""
        kpt = calc.wfs.kpt_qs[0][spin]
        gd = calc.wfs.gd.new_descriptor(pbc_c=np.ones(3, bool),
                                        comm=serial_comm)
        psit_nR = gd.empty(n2 - n1)
        for band, psit_R in enumerate(psit_nR):
            psit_R[:] = calc.get_pseudo_wave_function(
                band + n1,
                spin=spin) * Bohr**1.5

        return WaveFunctions(psit_nR,
                             kpt.projections.as_dict_on_master(n1, n2),
                             spin,
                             calc.setups,
                             gd=gd)

    def __len__(self) -> int:
        return len(self.psit_nR)


def create_compensation_charge(setups: List[Setup],
                               pd: PWDescriptor,
                               spos_ac: Array2D) -> PWLFC:
    compensation_charge = PWLFC([data.ghat_l for data in setups], pd)
    compensation_charge.set_positions(spos_ac)
    return compensation_charge


def zfs1(wf1: WaveFunctions,
         wf2: WaveFunctions,
         compensation_charge: PWLFC) -> Array2D:
    """Compute dipole coupling."""
    pd = wf1.pd
    setups = wf1.setups
    N2 = len(wf2)

    G_G = pd.G2_qG[0]**0.5
    G_G[0] = 1.0
    G_Gv = pd.get_reciprocal_vectors(add_q=False) / G_G[:, np.newaxis]

    n_sG = pd.zeros(2)
    for n_G, wf in zip(n_sG, [wf1, wf2]):
        D_aii = {}
        Q_aL = {}
        for a, P_ni in wf.P_ani.items():
            D_ii = np.einsum('ni, nj -> ij', P_ni, P_ni)
            D_aii[a] = D_ii
            Q_aL[a] = np.einsum('ij, ijL -> L', D_ii, setups[a].Delta_iiL)

        for psit_R in wf.psit_nR:
            n_G += pd.fft(psit_R**2)

        compensation_charge.add(n_G, Q_aL)

    nn_G = (n_sG[0] * n_sG[1].conj()).real
    D_vv = zfs2(pd, G_Gv, nn_G)

    n_nG = pd.empty(N2)
    for n1, psit1_R in enumerate(wf1.psit_nR):
        D_anii = {}
        Q_anL = {}
        for a, P1_ni in wf1.P_ani.items():
            D_nii = np.einsum('i, nj -> nij', P1_ni[n1], wf2.P_ani[a])
            D_anii[a] = D_nii
            Q_anL[a] = np.einsum('nij, ijL -> nL',
                                 D_nii, setups[a].Delta_iiL)

        for n_G, psit2_R in zip(n_nG, wf2.psit_nR):
            n_G[:] = pd.fft(psit1_R * psit2_R)

        compensation_charge.add(n_nG, Q_anL)

        nn_G = (n_nG * n_nG.conj()).sum(axis=0).real
        D_vv -= zfs2(pd, G_Gv, nn_G)

    D_vv -= np.trace(D_vv) / 3 * np.eye(3)  # should be traceless

    sign = 1.0 if wf1.spin == wf2.spin else -1.0

    return sign * alpha**2 * pi * D_vv * Ha


def zfs2(pd: PWDescriptor,
         G_Gv: Array2D,
         nn_G: Array1D) -> Array2D:
    """Integral."""
    D_vv = np.einsum('gv, gw, g -> vw', G_Gv, G_Gv, nn_G)
    D_vv *= 2 * pd.gd.dv / pd.gd.N_c.prod()
    return D_vv


def convert_tensor(D_vv: Array2D,
                   unit: str = 'eV') -> Tuple[float, float, Array1D, Array2D]:
    """Convert 3x3 tensor to D, E and easy axis.

    Input tensor must be in eV and the result can be returned in
    eV, μeV, MHz or 1/cm acording to the value uf *unit*
    (must be one of "eV", "ueV", "MHz", "1/cm").

    >>> D_vv = np.diag([1, 2, 3])
    >>> D, E, axis, _ = convert_tensor(D_vv)
    >>> D
    4.5
    >>> E
    0.5
    >>> axis
    array([0., 0., 1.])
    """
    if unit == 'ueV':
        scale = 1e6
    elif unit == 'MHz':
        scale = _e / _hplanck * 1e-6
    elif unit == '1/cm':
        scale = _e / _hplanck / _c / 100
    elif unit == 'eV':
        scale = 1.0
    else:
        raise ValueError(f'Unknown unit: {unit}')

    (e1, e2, e3), U = np.linalg.eigh(D_vv * scale)

    if abs(e1) > abs(e3):
        D = 1.5 * e1
        E = 0.5 * (e2 - e3)
        axis = U[:, 0]
    else:
        D = 1.5 * e3
        E = 0.5 * (e2 - e1)
        axis = U[:, 2]

    return D, E, axis, D_vv * scale


def main(argv: List[str] = None) -> Array2D:
    """CLI interface."""
    import argparse

    parser = argparse.ArgumentParser(
        prog='python3 -m gpaw.zero_field_splitting',
        description='...')
    add = parser.add_argument
    add('file', metavar='input-file',
        help='GPW-file with wave functions.')
    add('-u', '--unit', default='ueV', choices=['ueV', 'MHz', '1/cm'],
        help='Unit.  Must be "ueV" (micro-eV, default), "MHz" or "1/cm".')
    add('-m', '--method', type=int, default=1)

    args = parser.parse_intermixed_args(argv)

    calc = GPAW(args.file)

    D_vv = zfs(calc, args.method)
    D, E, axis, D_vv = convert_tensor(D_vv, args.unit)

    unit = args.unit
    if unit == 'ueV':
        unit = 'μeV'

    print('D_ij = (' +
          ',\n        '.join('(' + ', '.join(f'{d:10.3f}' for d in D_v) + ')'
                             for D_v in D_vv) +
          ') ', unit)
    print('i, j = x, y, z')
    print()
    print(f'D = {D:.3f} {unit}')
    print(f'E = {E:.3f} {unit}')
    x, y, z = axis
    print(f'axis = ({x:.3f}, {y:.3f}, {z:.3f})')

    return D_vv


if __name__ == '__main__':
    main()



from typing import Tuple

import numpy as np
import ase.units as units

from gpaw.typing import Array2D, ArrayLike2D
from gpaw.utilities import pack2, unpack2


def parse_hubbard_string(type: str) -> Tuple[str, 'HubbardU']:

    # Parse DFT+U parameters from type-string:
    # Examples: "type:l,U" or "type:l,U,scale"
    type, lus = type.split(':')
    if type == '':
        type = 'paw'

    l = []
    U = []
    scale = []

    for lu in lus.split(';'):  # Multiple U corrections
        l_, u_, scale_ = (lu + ',,').split(',')[:3]
        l.append('spdf'.find(l_))
        U.append(float(u_) / units.Hartree)
        if scale_:
            scale.append(bool(int(scale_)))
        else:
            scale.append(True)
    return type, HubbardU(U, l, scale)


class HubbardU:
    def __init__(self, U, l, scale=1):
        self.scale = scale
        self.U = U
        self.l = l

    def _tuple(self):
        # Tests use this method to compare to expected values
        return (self.l, self.U, self.scale)

    def calculate(self, setup, D_sp):
        e_xc = 0.0
        dH_sp = np.zeros_like(D_sp)
        for l, U, scale in zip(self.l, self.U, self.scale):
            e1_xc, dH1_sp = hubbard(
                setup.l_j, setup.lq, D_sp,
                l=l, U=U, scale=scale)

            e_xc += e1_xc
            dH_sp += dH1_sp
        return e_xc, dH_sp

    def descriptions(self):
        for U, l, scale in zip(self.U, self.l, self.scale):
            yield f'Hubbard: {{U: {U * units.Ha},  # eV\n'
            yield f'          l: {l},\n'
            yield f'          scale: {bool(scale)}}}'


def hubbard(l_j, lq,
            D_sp,
            l: int,
            U: float,
            scale: bool) -> Tuple[float, ArrayLike2D]:
    nspins = len(D_sp)

    nl = np.where(np.equal(l_j, l))[0]
    nn = (2 * np.array(l_j) + 1)[0:nl[0]].sum()

    e_xc = 0.0
    dH_sp = []

    s = 0
    for D_p in D_sp:
        N_mm, V = aoom(l_j, lq, unpack2(D_p), l, scale)
        N_mm = N_mm / 2 * nspins

        if nspins == 4:
            N_mm = N_mm / 2.0
            if s == 0:
                Eorb = U / 2. * (N_mm -
                                 0.5 * np.dot(N_mm, N_mm)).trace()

                Vorb = U / 2. * (np.eye(2 * l + 1) - N_mm)

            else:
                Eorb = U / 2. * (-0.5 * np.dot(N_mm, N_mm)).trace()

                Vorb = -U / 2. * N_mm
        else:
            Eorb = U / 2. * (N_mm -
                             np.dot(N_mm, N_mm)).trace()

            Vorb = U * (0.5 * np.eye(2 * l + 1) - N_mm)

        e_xc += Eorb
        if nspins == 1:
            # add contribution of other spin manyfold
            e_xc += Eorb

        if len(nl) == 2:
            mm = (2 * np.array(l_j) + 1)[0:nl[1]].sum()

            V[nn:nn + 2 * l + 1, nn:nn + 2 * l + 1] *= Vorb
            V[mm:mm + 2 * l + 1, nn:nn + 2 * l + 1] *= Vorb
            V[nn:nn + 2 * l + 1, mm:mm + 2 * l + 1] *= Vorb
            V[mm:mm + 2 * l + 1, mm:mm + 2 * l + 1] *= Vorb
        else:
            V[nn:nn + 2 * l + 1, nn:nn + 2 * l + 1] *= Vorb

        dH_sp.append(pack2(V))
        s += 1

    return e_xc, dH_sp


def aoom(l_j, lq,
         DM: Array2D,
         l: int,
         scale: bool = True) -> Tuple[Array2D, Array2D]:
    """Atomic Orbital Occupation Matrix.

    Determine the Atomic Orbital Occupation Matrix (aoom) for a
    given l-quantum number.

    This operation, takes the density matrix (DM), which for
    example is given by unpack2(D_asq[i][spin]), and corrects for
    the overlap between the selected orbitals (l) upon which the
    the density is expanded (ex <p|p*>,<p|p>,<p*|p*> ).

    Returned is only the "corrected" part of the density matrix,
    which represents the orbital occupation matrix for l=2 this is
    a 5x5 matrix.
    """
    nl = np.where(np.equal(l_j, l))[0]
    V = np.zeros(np.shape(DM))
    if len(nl) == 2:
        aa = nl[0] * len(l_j) - (nl[0] - 1) * nl[0] // 2
        bb = nl[1] * len(l_j) - (nl[1] - 1) * nl[1] // 2
        ab = aa + nl[1] - nl[0]

        if not scale:
            lq_a = lq[aa]
            lq_ab = lq[ab]
            lq_b = lq[bb]
        else:
            lq_a = 1
            lq_ab = lq[ab] / lq[aa]
            lq_b = lq[bb] / lq[aa]

        # and the correct entrances in the DM
        nn = (2 * np.array(l_j) + 1)[0:nl[0]].sum()
        mm = (2 * np.array(l_j) + 1)[0:nl[1]].sum()

        # finally correct and add the four submatrices of NC_DM
        A = DM[nn:nn + 2 * l + 1, nn:nn + 2 * l + 1] * lq_a
        B = DM[nn:nn + 2 * l + 1, mm:mm + 2 * l + 1] * lq_ab
        C = DM[mm:mm + 2 * l + 1, nn:nn + 2 * l + 1] * lq_ab
        D = DM[mm:mm + 2 * l + 1, mm:mm + 2 * l + 1] * lq_b

        V[nn:nn + 2 * l + 1, nn:nn + 2 * l + 1] = lq_a
        V[nn:nn + 2 * l + 1, mm:mm + 2 * l + 1] = lq_ab
        V[mm:mm + 2 * l + 1, nn:nn + 2 * l + 1] = lq_ab
        V[mm:mm + 2 * l + 1, mm:mm + 2 * l + 1] = lq_b

        return A + B + C + D, V
    else:
        assert len(nl) == 1
        assert l_j[-1] == l
        nn = (2 * np.array(l_j) + 1)[0:nl[0]].sum()
        A = DM[nn:nn + 2 * l + 1, nn:nn + 2 * l + 1] * lq[-1]
        V[nn:nn + 2 * l + 1, nn:nn + 2 * l + 1] = lq[-1]
        return A, V



import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.mixer import Mixer
from gpaw.test import equal
from gpaw.test import gen


@pytest.mark.ofdft
@pytest.mark.libxc
def test_ofdft_ofdft(in_tmp_dir):
    a = 6.0
    c = a / 2
    # d = 1.8

    elements = ['C', 'Be']
    results = [0.243, 9.9773]
    electrons = [6, 3]
    charges = [0, 1]
    xcname = '1.0_LDA_K_TF+1.0_LDA_X'

    setups = {}
    for symbol in elements:
        s = gen(symbol, xcname=xcname, scalarrel=False, orbital_free=True,
                gpernode=75)
        setups[symbol] = s

    for element, result, e, charge in zip(elements,
                                          results,
                                          electrons,
                                          charges):
        atom = Atoms(element,
                     positions=[(c, c, c)],
                     cell=(a, a, a))

        mixer = Mixer(0.3, 5, 1)
        calc = GPAW(gpts=(32, 32, 32),
                    txt='-',
                    xc=xcname,
                    setups=setups,
                    eigensolver='cg', mixer=mixer, charge=charge)

        atom.calc = calc

        E = atom.get_total_energy()
        n = calc.get_all_electron_density()

        dv = atom.get_volume() / calc.get_number_of_grid_points().prod()
        I = n.sum() * dv / 2**3

        equal(I, e, 1.0e-6)
        equal(result, E, 1.0e-2)



# This tests calculates the force on the atoms of a
# slightly distorted Silicon primitive cell.
#
# If the test fails, set the fd boolean below to enable a (costly) finite
# difference check.

import numpy as np
import pytest
from ase import Atoms

from gpaw import GPAW
from gpaw.atom.basis import BasisMaker
from gpaw.test import equal


@pytest.mark.later
def test_generic_guc_force():
    sibasis = BasisMaker('Si').generate(
        2, 1, energysplit=0.3, tailnorm=0.03**.5)
    basis = {'Si': sibasis}

    a = 5.475
    system = Atoms(symbols='Si2', pbc=True,
                   cell=0.5 * a * np.array([(1, 1, 0),
                                            (1, 0, 1),
                                            (0, 1, 1)]),
                   scaled_positions=[(0.0, 0.0, 0.0),
                                     (0.23, 0.23, 0.23)])

    calc = GPAW(h=0.2,
                mode='lcao',
                basis=basis,
                kpts=(2, 2, 2),
                convergence={'density': 1e-5, 'energy': 1e-6}
                )
    system.calc = calc

    F_ac = system.get_forces()

    # Previous FD result, generated by disabled code below
    F_ac_ref = np.array(
        [[-1.3967114867039498, -1.3967115816022613, -1.396711581510779],
         [1.397400325299003, 1.3974003410455182, 1.3974003410801572]])

    err_ac = np.abs(F_ac - F_ac_ref)
    err = err_ac.max()

    print('Force')
    print(F_ac)
    print()
    print('Reference result')
    print(F_ac_ref)
    print()
    print('Error')
    print(err_ac)
    print()
    print('Max error')
    print(err)

    # ASE uses dx = [+|-] 0.001 by default,
    # error should be around 2e-3.  In fact 4e-3 would probably be acceptable

    equal(err, 0, 6e-3)

    # Set boolean to run new FD check
    fd = False

    if fd:
        from ase.calculators.test import numeric_force
        system.calc = calc.new(symmetry='off')
        F_ac_fd = [[numeric_force(system, a, i) for i in range(3)]
                   for a in range(2)]
        print('Self-consistent forces')
        print(F_ac)
        print('FD')
        print(F_ac_fd)
        print(repr(F_ac_fd))
        print(F_ac - F_ac_fd, np.abs(F_ac - F_ac_fd).max())

        assert np.abs(F_ac - F_ac_fd).max() < 4e-3



import pytest
from gpaw.mpi import world
from ase import Atoms
from gpaw import GPAW
from gpaw.spinorbit import soc_eigenstates

pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


@pytest.mark.soc
def test_spinorbit_Kr():
    a = Atoms('Kr')
    a.center(vacuum=3.0)

    calc = GPAW(mode='pw', xc='LDA')

    a.calc = calc
    a.get_potential_energy()

    e_n = calc.get_eigenvalues()
    e_m = soc_eigenstates(calc)[0].eig_m

    assert e_n[0] - e_m[0] == pytest.approx(0.0, abs=1.0e-3)
    assert e_n[1] - e_m[2] == pytest.approx(0.452, abs=1.0e-3)
    assert e_n[2] - e_m[4] == pytest.approx(-0.226, abs=1.0e-3)



from __future__ import annotations
from math import pi

import _gpaw
import gpaw.gpu.kernels as gpu_kernels
import numpy as np
from gpaw.core.atom_arrays import AtomArraysLayout, AtomDistribution
from gpaw.core.atom_centered_functions import AtomCenteredFunctions
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.gpu import cupy_is_fake
from gpaw.lfc import BaseLFC
from gpaw.new import prod
from gpaw.pw.lfc import ft
from gpaw.spherical_harmonics import Y, nablarlYL
from gpaw.utilities.blas import mmm
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from gpaw.core.plane_waves import PlaneWaves


class PlaneWaveAtomCenteredFunctions(AtomCenteredFunctions):
    def __init__(self,
                 functions,
                 fracpos,
                 pw,
                 atomdist=None,
                 xp=None):
        AtomCenteredFunctions.__init__(self, functions, fracpos, atomdist)
        self.pw = pw
        self.xp = xp or np

    def _lazy_init(self):
        if self._lfc is not None:
            return

        self._lfc = PWLFC(self.functions, self.pw, xp=self.xp)

        if self._atomdist is None:
            self._atomdist = AtomDistribution.from_number_of_atoms(
                len(self.fracpos_ac), self.pw.comm)
        else:
            assert self.pw.comm is self._atomdist.comm

        self._lfc.set_positions(self.fracpos_ac, self._atomdist)
        self._layout = AtomArraysLayout([sum(2 * f.l + 1 for f in funcs)
                                         for funcs in self.functions],
                                        self._atomdist,
                                        self.pw.dtype,
                                        xp=self.xp)

    def __repr__(self):
        s = super().__repr__()
        if self.xp is np:
            return s
        return s[:-1] + ', xp=cp)'

    def to_uniform_grid(self,
                        out: UniformGridFunctions,
                        scale: float = 1.0) -> UniformGridFunctions:
        out_G = self.pw.zeros(xp=out.xp)
        self.add_to(out_G, scale)
        return out_G.ifft(out=out)


class PWLFC(BaseLFC):
    def __init__(self,
                 functions,
                 pw: PlaneWaves,
                 blocksize=5000, *, xp):
        """Reciprocal-space plane-wave localized function collection.

        spline_aj: list of list of spline objects
            Splines.
        pd: PWDescriptor
            Plane-wave descriptor object.
        blocksize: int
            Block-size to use when looping over G-vectors.  Use None for
            doing all G-vectors in one big block.
        """

        self.xp = xp
        self.pw = pw
        self.spline_aj = functions

        self.dtype = pw.dtype

        self.initialized = False

        # These will be filled in later:
        self.Y_GL = None
        self.emiGR_Ga = None
        self.f_Gs = None
        self.l_s = None
        self.a_J = None
        self.s_J = None
        self.lmax = None

        if blocksize is not None:
            if pw.maxmysize <= blocksize:
                # No need to block G-vectors
                blocksize = None
        self.blocksize = blocksize

        # These are set later in set_potitions():
        self.eikR_a = None
        self.my_atom_indices = None
        self.my_indices = None
        self.pos_av = None
        self.nI = None

        self.comm = pw.comm

    def initialize(self):
        """Initialize position-independent stuff."""
        if self.initialized:
            return

        xp = self.xp

        splines = {}  # Dict[Spline, int]
        for spline_j in self.spline_aj:
            for spline in spline_j:
                if spline not in splines:
                    splines[spline] = len(splines)
        nsplines = len(splines)

        nJ = sum(len(spline_j) for spline_j in self.spline_aj)

        self.f_Gs = xp.empty(self.pw.myshape + (nsplines,))
        self.l_s = np.empty(nsplines, np.int32)
        self.a_J = np.empty(nJ, np.int32)
        self.s_J = np.empty(nJ, np.int32)
        self.I_J = np.empty(nJ, np.int32)
        # Fourier transform radial functions:
        J = 0
        done = set()  # Set[Spline]
        I = 0
        for a, spline_j in enumerate(self.spline_aj):
            for spline in spline_j:
                s = splines[spline]  # get spline index
                if spline not in done:
                    f = ft(spline)
                    G_G = (2 * self.pw.ekin_G)**0.5
                    self.f_Gs[:, s] = xp.asarray(f.map(G_G))
                    self.l_s[s] = spline.get_angular_momentum_number()
                    done.add(spline)
                self.a_J[J] = a
                self.s_J[J] = s
                self.I_J[J] = I
                I += 2 * spline.get_angular_momentum_number() + 1
                J += 1

        self.lmax = max(self.l_s, default=-1)

        # Spherical harmonics:
        G_Gv = self.pw.G_plus_k_Gv
        self.Y_GL = xp.empty((len(G_Gv), (self.lmax + 1)**2))
        for L in range((self.lmax + 1)**2):
            self.Y_GL[:, L] = xp.asarray(Y(L, *G_Gv.T))

        self.l_s = xp.asarray(self.l_s)
        self.a_J = xp.asarray(self.a_J)
        self.s_J = xp.asarray(self.s_J)
        self.I_J = xp.asarray(self.I_J)

        self.initialized = True

    def get_function_count(self, a):
        return sum(2 * spline.get_angular_momentum_number() + 1
                   for spline in self.spline_aj[a])

    def set_positions(self, spos_ac, atomdist):
        self.initialize()

        xp = self.xp

        if self.pw.dtype == float:
            self.eikR_a = xp.ones(len(spos_ac))
        else:
            self.eikR_a = xp.asarray(
                np.exp(2j * pi * (spos_ac @ self.pw.kpt_c)))

        self.pos_av = np.dot(spos_ac, self.pw.cell)

        Gk_Gv = xp.asarray(self.pw.G_plus_k_Gv)
        GkR_Ga = Gk_Gv @ xp.asarray(self.pos_av.T)
        self.emiGR_Ga = xp.exp(-1j * GkR_Ga) * self.eikR_a

        rank_a = atomdist.rank_a

        self.my_atom_indices = []
        self.my_indices = []
        I1 = 0
        for a, rank in enumerate(rank_a):
            I2 = I1 + self.get_function_count(a)
            if rank == self.comm.rank:
                self.my_atom_indices.append(a)
                self.my_indices.append((a, I1, I2))
            I1 = I2
        self.nI = I1

    def expand(self, G1=0, G2=None, cc=False):
        """Expand functions in plane-waves.

        q: int
            k-point index.
        G1: int
            Start G-vector index.
        G2: int
            End G-vector index.
        cc: bool
            Complex conjugate.
        """
        xp = self.xp

        if G2 is None:
            G2 = self.Y_GL.shape[0]

        emiGR_Ga = self.emiGR_Ga[G1:G2]
        f_Gs = self.f_Gs[G1:G2]
        Y_GL = self.Y_GL[G1:G2]

        if self.dtype == complex:
            f_GI = xp.empty((G2 - G1, self.nI), complex)
        else:
            # Special layout because BLAS does not have real-complex
            # multiplications.  f_GI(G,I) layout:
            #
            #    real(G1, 0),   real(G1, 1),   ...
            #    imag(G1, 0),   imag(G1, 1),   ...
            #    real(G1+1, 0), real(G1+1, 1), ...
            #    imag(G1+1, 0), imag(G1+1, 1), ...
            #    ...

            f_GI = xp.empty((2 * (G2 - G1), self.nI))

        if xp is np:
            # Fast C-code:
            _gpaw.pwlfc_expand(f_Gs, emiGR_Ga, Y_GL,
                               self.l_s, self.a_J, self.s_J,
                               cc, f_GI)
            return f_GI
        elif cupy_is_fake or getattr(_gpaw, 'gpu_aware_mpi', False):
            gpu_kernels.pwacf_expand(f_Gs, emiGR_Ga, Y_GL,
                                     self.l_s, self.a_J, self.s_J,
                                     cc, f_GI, self.I_J)
            return f_GI

        # Equivalent slow Python code:
        f_GI = xp.empty((G2 - G1, self.nI), complex)
        I1 = 0
        for J, (a, s) in enumerate(zip(self.a_J, self.s_J)):
            l = self.l_s[s]
            I2 = I1 + 2 * l + 1
            f_GI[:, I1:I2] = (f_Gs[:, s] *
                              emiGR_Ga[:, a] *
                              Y_GL[:, l**2:(l + 1)**2].T *
                              (-1.0j)**l).T
            I1 = I2
        if cc:
            f_GI = f_GI.conj()
        if self.dtype == float:
            f_GI = f_GI.T.copy().view(float).T.copy()

        return f_GI

    def block(self, ensure_same_number_of_blocks=False):
        nG = self.Y_GL.shape[0]
        B = self.blocksize
        if B:
            G1 = 0
            while G1 < nG:
                G2 = min(G1 + B, nG)
                yield G1, G2
                G1 = G2
            if ensure_same_number_of_blocks:
                # Make sure we yield the same number of times:
                nb = (self.pw.maxmysize + B - 1) // B
                mynb = (nG + B - 1) // B
                if mynb < nb:
                    yield nG, nG  # empty block
        else:
            yield 0, nG

    def add(self, a_xG, c_axi=1.0, q=None):
        if self.nI == 0:
            return
        c_xI = self.xp.empty(a_xG.shape[:-1] + (self.nI,), self.dtype)

        if isinstance(c_axi, float):
            assert a_xG.ndim == 1
            c_xI[:] = c_axi
        else:
            if self.comm.size != 1:
                c_xI[:] = 0.0
            for a, I1, I2 in self.my_indices:
                c_xI[..., I1:I2] = c_axi[a] * self.eikR_a[a].conj()
            if self.comm.size != 1:
                self.comm.sum(c_xI)

        nx = prod(c_xI.shape[:-1])
        c_xI = c_xI.reshape((nx, self.nI))
        a_xG = a_xG.reshape((nx, a_xG.shape[-1])).view(self.dtype)

        for G1, G2 in self.block():
            f_GI = self.expand(G1, G2, cc=False)

            if self.dtype == float:
                # f_IG = f_IG.view(float)
                G1 *= 2
                G2 *= 2

            if self.xp is np:
                mmm(1.0 / self.pw.dv, c_xI, 'N', f_GI, 'T',
                    1.0, a_xG[:, G1:G2])
            else:
                self.xp.cublas.gemm('N', 'T',
                                    c_xI, f_GI, a_xG[:, G1:G2],
                                    1.0 / self.pw.dv, 1.0)

    def integrate(self, a_xG, c_axi=None, q=-1):
        xp = self.xp
        if self.nI == 0:
            return c_axi
        c_xI = xp.zeros(a_xG.shape[:-1] + (self.nI,), self.dtype)

        nx = prod(c_xI.shape[:-1])
        b_xI = c_xI.reshape((nx, self.nI))
        a_xG = a_xG.reshape((nx, a_xG.shape[-1]))

        alpha = 1.0  # / self.pd.gd.N_c.prod()
        if self.dtype == float:
            alpha *= 2
            a_xG = a_xG.view(float)

        if c_axi is None:
            c_axi = self.dict(a_xG.shape[:-1])

        x = 0.0
        for G1, G2 in self.block():
            f_GI = self.expand(G1, G2, cc=self.dtype == complex)
            if self.dtype == float:
                if G1 == 0 and self.comm.rank == 0:
                    f_GI[0] *= 0.5
                G1 *= 2
                G2 *= 2
            if xp is np:
                mmm(alpha, a_xG[:, G1:G2], 'N', f_GI, 'N', x, b_xI)
            else:
                xp.cublas.gemm('N', 'N',
                               a_xG[:, G1:G2], f_GI, b_xI,
                               alpha, x)
            x = 1.0

        self.comm.sum(b_xI)
        for a, I1, I2 in self.my_indices:
            c_axi[a][:] = self.eikR_a[a] * c_xI[..., I1:I2]

        return c_axi

    def derivative(self, a_xG, c_axiv=None, q=-1):
        xp = self.xp
        c_vxI = xp.zeros((3,) + a_xG.shape[:-1] + (self.nI,), self.dtype)
        nx = prod(c_vxI.shape[1:-1])
        b_vxI = c_vxI.reshape((3, nx, self.nI))
        a_xG = a_xG.reshape((nx, a_xG.shape[-1])).view(self.dtype)

        alpha = 1.0

        if c_axiv is None:
            c_axiv = self.dict(a_xG.shape[:-1], derivative=True)

        x = 0.0
        for G1, G2 in self.block():
            f_GI = self.expand(G1, G2, cc=True)
            G_Gv = xp.asarray(self.pw.G_plus_k_Gv[G1:G2])
            if self.dtype == float:
                d_GI = xp.empty(f_GI.shape)
                for v in range(3):
                    d_GI[::2] = f_GI[1::2] * G_Gv[:, v, np.newaxis]
                    d_GI[1::2] = f_GI[::2] * G_Gv[:, v, np.newaxis]
                    if xp is np:
                        mmm(2 * alpha,
                            a_xG[:, 2 * G1:2 * G2], 'N',
                            d_GI, 'N',
                            x, b_vxI[v])
                    else:
                        xp.cublas.gemm('N', 'N',
                                       a_xG[:, 2 * G1:2 * G2],
                                       d_GI,
                                       b_vxI[v],
                                       2 * alpha, x)
            else:
                for v in range(3):
                    if xp is np:
                        mmm(-alpha,
                            a_xG[:, G1:G2], 'N',
                            f_GI * G_Gv[:, v, np.newaxis], 'N',
                            x, b_vxI[v])
                    else:
                        xp.cublas.gemm('N', 'N',
                                       a_xG[:, G1:G2],
                                       f_GI * G_Gv[:, v, np.newaxis],
                                       b_vxI[v],
                                       -alpha, x)
            x = 1.0

        self.comm.sum(c_vxI)

        for v in range(3):
            if self.dtype == float:
                for a, I1, I2 in self.my_indices:
                    c_axiv[a][..., v] = c_vxI[v, ..., I1:I2]
            else:
                for a, I1, I2 in self.my_indices:
                    c_axiv[a][..., v] = (1.0j * self.eikR_a[a] *
                                         c_vxI[v, ..., I1:I2])

        return c_axiv

    def stress_tensor_contribution(self, a_xG, c_axi=1.0):
        xp = self.xp
        cache = {}
        things = []
        I1 = 0
        lmax = 0
        for a, spline_j in enumerate(self.spline_aj):
            for spline in spline_j:
                if spline not in cache:
                    s = ft(spline)
                    G_G = (2 * self.pw.ekin_G)**0.5
                    f_G = []
                    dfdGoG_G = []
                    for G in G_G:
                        f, dfdG = s.get_value_and_derivative(G)
                        if G < 1e-10:
                            G = 1.0
                        f_G.append(f)
                        dfdGoG_G.append(dfdG / G)
                    f_G = xp.array(f_G)
                    dfdGoG_G = xp.array(dfdGoG_G)
                    cache[spline] = (f_G, dfdGoG_G)
                else:
                    f_G, dfdGoG_G = cache[spline]
                l = spline.l
                lmax = max(l, lmax)
                I2 = I1 + 2 * l + 1
                things.append((a, l, I1, I2, f_G, dfdGoG_G))
                I1 = I2

        if isinstance(c_axi, float):
            c_axi = dict((a, c_axi) for a in range(len(self.pos_av)))

        G0_Gv = self.pw.G_plus_k_Gv

        stress_vv = xp.zeros((3, 3))
        for G1, G2 in self.block(ensure_same_number_of_blocks=True):
            G_Gv = G0_Gv[G1:G2]
            Z_LvG = xp.array([nablarlYL(L, G_Gv.T)
                              for L in range((lmax + 1)**2)])
            G_Gv = xp.asarray(G_Gv)
            aa_xG = a_xG[..., G1:G2]
            for v1 in range(3):
                for v2 in range(3):
                    stress_vv[v1, v2] += self._stress_tensor_contribution(
                        v1, v2, things, G1, G2, G_Gv, aa_xG, c_axi, Z_LvG)

        self.comm.sum(stress_vv)

        return stress_vv

    def _stress_tensor_contribution(self, v1, v2, things, G1, G2,
                                    G_Gv, a_xG, c_axi, Z_LvG):
        xp = self.xp
        f_IG = xp.empty((self.nI, G2 - G1), complex)
        emiGR_Ga = self.emiGR_Ga[G1:G2]
        Y_LG = self.Y_GL.T
        for a, l, I1, I2, f_G, dfdGoG_G in things:
            L1 = l**2
            L2 = (l + 1)**2
            f_IG[I1:I2] = (emiGR_Ga[:, a] * (-1.0j)**l *
                           (dfdGoG_G[G1:G2] * G_Gv[:, v1] * G_Gv[:, v2] *
                            Y_LG[L1:L2, G1:G2] +
                            f_G[G1:G2] * G_Gv[:, v1] * Z_LvG[L1:L2, v2]))

        c_xI = xp.zeros(a_xG.shape[:-1] + (self.nI,), self.pw.dtype)

        x = prod(c_xI.shape[:-1])
        b_xI = c_xI.reshape((x, self.nI))
        a_xG = a_xG.reshape((x, a_xG.shape[-1]))

        alpha = 1.0
        if self.pw.dtype == float:
            alpha = 2.0
            if G1 == 0 and self.pw.comm.rank == 0:
                f_IG[:, 0] *= 0.5
            f_IG = f_IG.view(float)
            a_xG = a_xG.copy().view(float)

        if xp is np:
            mmm(alpha, a_xG, 'N', f_IG, 'C', 0.0, b_xI)
        else:
            xp.cublas.gemm('N', 'H', a_xG, f_IG, b_xI, alpha, 0.0)
        self.comm.sum(b_xI)

        stress = 0.0
        for a, I1, I2 in self.my_indices:
            stress -= self.eikR_a[a] * (c_axi[a] * c_xI[..., I1:I2]).sum()
        return stress.real
